#!/usr/bin/env python3

import argparse
import numpy as np
import odil
from odil.runtime import mod


def test(method, ndim, loc):
    cshapeh = 3 + np.array(range(ndim))
    cshape = cshapeh * 2

    dimnames = ['x', 'y', 'z', 'w'][:ndim]
    domain = odil.Domain(cshape=cshape, dimnames=dimnames)
    domainh = odil.Domain(cshape=cshapeh, dimnames=dimnames)
    dtype = domain.dtype
    xx = domain.points(loc=loc)
    xxh = domainh.points(loc=loc)

    def func(xx):
        # Linear function in space.
        return sum(x * np.sqrt(i + 1) for i, x in enumerate(xx))

    u = func(xx)
    uh = func(xxh)
    ui = odil.core.interp_to_finer(uh, loc=loc, mod=mod, method=method)
    error = mod.max(abs(ui - u))
    msg = '{:.6e}'.format(error)
    failed = error > np.finfo(dtype).eps * 100
    return failed, msg


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(  #
        '--method', type=str, nargs='+', default=['conv', 'stack'])
    parser.add_argument('--ndim', type=int, nargs='+', default=[1, 2, 3, 4])
    parser.add_argument('--loc',
                        type=str,
                        nargs='+',
                        default=['cccc', 'nnnn', 'cnnn', 'nccc'])
    args = parser.parse_args()
    count_failed = 0
    for method in args.method:
        for ndim in args.ndim:
            for loc in {s[:ndim] for s in args.loc}:
                # Skip unsupported combinations.
                if ndim == 4 and method == 'conv':
                    continue
                try:
                    failed, msg = test(method, ndim, loc)
                except Exception as e:
                    msg = e
                count_failed += bool(failed)
                print(f'method={method:>5}, ndim={ndim}, loc={loc:>4}:',
                      "{} {}".format(msg, 'FAIL' if failed else 'PASS'))
    exit(count_failed)


if __name__ == "__main__":
    main()



import sys
import os
import json
import argparse
import psutil
import time
import numpy as np

from .optimizer import make_optimizer, Optimizer
from .history import History
g_log_file = sys.stderr  # File used by printlog()
g_log_echo = False  # True if printlog() should print to stderr.


def assert_equal(first, second, msg=None):
    if not (first == second):
        raise ValueError("Expected equal '{:}' and '{:}'{}".format(
            first, second, msg))


def set_log_file(f=None, echo=None):
    global g_log_file, g_log_echo
    if f is not None:
        g_log_file = f
    if echo is not None:
        g_log_echo = echo


def printlog(*msg):
    m = ' '.join(map(str, msg)) + '\n'
    if g_log_echo and g_log_file != sys.stderr:
        sys.stderr.write(m)
        sys.stderr.flush()
    g_log_file.write(m)
    g_log_file.flush()


class Timer():

    def __init__(self):
        self._starts = []  # Stack of pairs (key, time).
        self.counters = dict()  # Key to time.

    def push(self, key=None):
        self._starts.append((key, time.time()))

    def pop(self, key=None):
        start = self._starts.pop()
        assert start[0] is None or key is None or start[0] == key, \
                "Inconsistent keys passed to push() and pop(): "\
                "{:} and {:}".format(start[0], key)
        if key is None:
            key = start[0]
        dt = time.time() - start[1]
        self.counters[key] = self.counters.get(key, 0.) + dt

    def append(self, timer):
        for k in timer.counters:
            self.counters[k] = self.counters.get(k, 0.) + timer.counters[k]


def get_error(u, v):
    e1 = np.mean(abs(u - v))
    e2 = np.mean((u - v)**2)**0.5
    einf = np.max(abs(u - v))
    return e1, e2, einf


def add_arguments(parser):
    parser.add_argument('--epochs',
                        type=int,
                        default=None,
                        help="Maximum epochs, "
                        "defaults to product of plot_every and frames")
    parser.add_argument('--every_factor',
                        type=float,
                        default=1,
                        help="Multiplier for all *_every options")
    parser.add_argument('--plot_every',
                        type=int,
                        default=5,
                        help="Epochs between plots")
    parser.add_argument('--report_every',
                        type=int,
                        default=10,
                        help="Epochs between reports to stdout")
    parser.add_argument('--history_every',
                        type=int,
                        default=1,
                        help="Epochs between entries of training history")
    parser.add_argument('--checkpoint_every',
                        type=int,
                        default=0,
                        help="Epochs between checkpoints")
    parser.add_argument('--frames',
                        type=int,
                        default=10,
                        help="Frames to plot. Zero disables first frame.")
    parser.add_argument('--outdir',
                        type=str,
                        default='.',
                        help='Output directory')
    parser.add_argument('--optimizer',
                        type=str,
                        default='adamn',
                        help="Optimizer")
    parser.add_argument('--seed',
                        default=1000,
                        type=int,
                        help="Seed for numpy.random and tensorflow.random")
    parser.add_argument('--plot_title',
                        type=int,
                        default=0,
                        help="Enable title in plots")
    parser.add_argument('--plotext',
                        type=str,
                        default='pdf',
                        help="Extension of plots")
    parser.add_argument('--history_full',
                        type=int,
                        default=0,
                        help="Number of epochs to write "
                        "history at every point")
    parser.add_argument('--montage',
                        type=int,
                        default=1,
                        help="Run montage after plotting")
    parser.add_argument('--double',
                        type=int,
                        default=None,
                        help="Double precision. Defaults to runtime.dtype")
    parser.add_argument('--epoch_start',
                        type=int,
                        default=0,
                        help="Initial value of epoch")
    parser.add_argument('--frame_start',
                        type=int,
                        default=0,
                        help="Initial value of frame")
    parser.add_argument('--checkpoint',
                        type=str,
                        help="Continue from checkpoint in state_*.pickle")
    parser.add_argument('--checkpoint_train',
                        type=str,
                        help="Continue from history in state_*_train.pickle"
                        ". By default, infers the name from --checkpoint"
                        ". Set to '' to disable default behavior")
    parser.add_argument('--callback_update_state',
                        type=int,
                        default=0,
                        help="Update state after callback")
    parser.add_argument('--bfgs_m',
                        type=int,
                        default=50,
                        help="History size for L-BFGS")
    parser.add_argument('--adam_epsilon',
                        type=float,
                        help="Parameter epsilon in Adam")
    parser.add_argument('--adam_beta_1',
                        type=float,
                        help="Parameter beta_1 in Adam")
    parser.add_argument('--adam_beta_2',
                        type=float,
                        help="Parameter beta_2 in Adam")
    parser.add_argument('--multigrid',
                        type=int,
                        default=0,
                        help="Use multigrid decomposition")
    parser.add_argument('--mg_interp',
                        type=str,
                        default='stack',
                        choices=[
                            'conv',
                            'stack',
                        ],
                        help="Multigrid interpolation method")
    parser.add_argument('--dump_data',
                        type=int,
                        default=1,
                        help="Dump data_*.pickle with every plot")
    parser.add_argument('--jac_nsmp0',
                        type=int,
                        default=50,
                        help="Number of samples "
                        "for initialization of Jacobi optimizer")
    parser.add_argument('--jac_nsmp1',
                        type=int,
                        default=1,
                        help="Number of samples "
                        "for each step of Jacobi optimizer")
    parser.add_argument('--jac_factor',
                        type=float,
                        default=1,
                        help="Factor for the diagonal update"
                        "for each step of Jacobi optimizer. "
                        "Increase above 1 for more weight to recent values")
    parser.add_argument('--jac_epsilon',
                        type=float,
                        default=1e-8,
                        help="Parameter epsilon in Jacobi optimizer. "
                        "Added to the diagonal to avoid division by zero")
    parser.add_argument('--nn_initializer',
                        type=str,
                        default='legacy',
                        choices=['legacy', 'glorot', 'lecun', 'he'],
                        help="Initializer for weights of neural networks")


def optimize_newton(args, problem, state, callback=None, **kwargs):
    domain = problem.domain
    mod = domain.mod

    def eval_pinfo(state):
        loss, _, terms, names, norms = problem.eval_loss_grad(state)
        pinfo = {'terms': terms, 'names': names, 'norms': norms, 'loss': loss}
        return pinfo

    from .linsolver import solve
    opt = Optimizer(name='newton', displayname='Newton')
    printlog("Running {} optimizer".format(opt.displayname))

    # Compute loss and residuals with initial state, to be used by callback.
    pinfo = eval_pinfo(state)
    if callback:
        callback(state, args.epoch_start, pinfo)

    for epoch in range(args.epoch_start, args.epochs):
        vector, matrix = problem.linearize(state)
        opt.evals += 1
        linstatus = dict()
        delta = solve(matrix, -vector, args, linstatus, args.linsolver)
        if args.linsolver_verbose:
            printlog(linstatus)
        packed = domain.pack_state(state)
        domain.unpack_state(packed + delta, state)
        if callback:
            pinfo = eval_pinfo(state)
            pinfo['linsolver'] = linstatus
            callback(state, epoch + 1, pinfo)


def optimize_grad(args, optname, problem, state, callback=None, **kwargs):
    domain = problem.domain
    mod = domain.mod

    def loss_grad(arrays):
        domain.arrays_to_state(arrays, state)
        loss, grads, terms, names, norms = problem.eval_loss_grad(state)
        pinfo = {'terms': terms, 'names': names, 'norms': norms, 'loss': loss}
        return loss, grads, pinfo

    def callback_wrap(arrays, epoch, pinfo):
        domain.arrays_to_state(arrays, state)
        callback(state, epoch, pinfo)
        if args.callback_update_state:
            new = domain.arrays_from_state(state)
            for i in range(len(new)):
                arrays[i] = new[i]

    # Custom parameters.
    if args.bfgs_m is not None:
        kwargs['m'] = args.bfgs_m
    if args.adam_epsilon is not None:
        kwargs['epsilon'] = args.adam_epsilon
    if args.adam_beta_1 is not None:
        kwargs['beta_1'] = args.adam_beta_1
    if args.adam_beta_2 is not None:
        kwargs['beta_2'] = args.adam_beta_2

    opt = make_optimizer(optname, dtype=domain.dtype, mod=mod, **kwargs)
    printlog("Running {} optimizer".format(opt.displayname))

    # Compute loss and residuals with initial state, to be used by callback.
    arrays = domain.arrays_from_state(state)
    _, _, pinfo = loss_grad(arrays)
    if callback:
        callback(state, args.epoch_start, pinfo)

    arrays, optinfo = opt.run(arrays,
                              loss_grad=loss_grad,
                              epochs=args.epochs - args.epoch_start,
                              callback=callback_wrap if callback else None,
                              epoch_start=args.epoch_start,
                              lr=args.lr,
                              **kwargs)
    return arrays, optinfo


def optimize(args, optname, problem, state, callback, **kwargs):
    if optname == 'newton':
        return optimize_newton(args, problem, state, callback, **kwargs)
    return optimize_grad(args, optname, problem, state, callback, **kwargs)


def get_memory_usage_kb():
    '''
    Returns current memory usage in KiB.
    '''
    process = psutil.Process()
    return process.memory_info().rss // 1024


def get_env_config():
    keys = [
        'OMP_NUM_THREADS', 'CUDA_VISIBLE_DEVICES', 'ODIL_WARN', 'ODIL_BACKEND',
        'ODIL_JIT', 'ODIL_MT', 'ODIL_DTYPE'
    ]
    return {k: os.environ.get(k, '') for k in keys}


def setup_outdir(args, relpath_args=None):
    '''
    Creates the output directory, configuration `args.json`, and log file `train.log`.
    Updates the arguments with new relative paths and number of epochs.
    Sets random seeds.

    args: `argparse.Namespace`
        Arguments produced by `argparse.ArgumentParser.parse_args()`.
    relpath_args: `list` of `str`
        List of names of attributes of `args` to be treated as relative paths.
        They are updated to paths relative to the output directory.
    '''
    from . import runtime
    mod = runtime.mod
    outdir = args.outdir
    os.makedirs(outdir, exist_ok=True)
    with open(os.path.join(outdir, 'args.json'), 'w') as f:
        d = dict(
            vars(args),
            **get_env_config(),
            runtime_backend=runtime.backend,
            runtime_dtype=runtime.dtype_name,
            runtime_jit=runtime.jit,
        )
        json.dump(d, f, sort_keys=True, indent=4)

    # Switch to output directory.
    os.makedirs(outdir, exist_ok=True)
    os.chdir(outdir)
    set_log_file(open("train.log", 'w'))

    # Update relative paths.
    if relpath_args is None:
        relpath_args = []
    for k in relpath_args:
        if getattr(args, k):
            setattr(args, k, os.path.relpath(getattr(args, k), start=outdir))

    # Update arguments.
    mulint = lambda v, k: None if v is None else max(1, round(v * k))
    args.plot_every = mulint(args.plot_every, args.every_factor)
    args.history_every = mulint(args.history_every, args.every_factor)
    args.report_every = mulint(args.report_every, args.every_factor)
    if args.epochs is None:
        args.epochs = args.frames * args.plot_every

    if args.seed is not None:
        np.random.seed(args.seed)
        mod.random.set_seed(args.seed)
    printlog(' '.join(sys.argv))


def make_callback(problem,
                  args=None,
                  epoch_func=None,
                  report_func=None,
                  history_func=None,
                  checkpoint_func=None,
                  plot_func=None):
    # Persistent state of the callback.
    cbinfo = argparse.Namespace()
    cbinfo.walltime = 0  # Walltime of the last call with task_report=True.
    cbinfo.epoch = 0  # Epoch of the last call with task_report=True.
    cbinfo.time_callback = 0  # Total time spent in callback.
    cbinfo.time_start = time.time()
    cbinfo.problem = problem
    cbinfo.args = args
    cbinfo.frame = 0

    if args.history_every:
        cbinfo.history = History(csvpath='train.csv', warmup=1)
    else:
        cbinfo.history = None

    def callback(state, epoch, pinfo):
        problem = cbinfo.problem
        domain = problem.domain
        args = cbinfo.args
        history = cbinfo.history
        time_prev = time.time()

        cbinfo.task_report = (args.report_every
                              and epoch % args.report_every == 0)
        cbinfo.task_history = (history is not None
                               and (epoch % args.history_every == 0
                                    or epoch < args.history_full))
        cbinfo.task_plot = (epoch % args.plot_every == 0
                            and (epoch or args.frames))
        cbinfo.task_checkpoint = (args.checkpoint_every
                                  and epoch % args.checkpoint_every == 0)

        cbinfo.pinfo = pinfo

        if isinstance(problem.tracers, dict):
            problem.tracers['epoch'] = epoch
        if epoch_func is not None:
            epoch_func(problem, state, epoch, cbinfo)

        # Subtract first part of callback time.
        curtime = time.time()
        cbinfo.time_callback += curtime - time_prev
        time_prev = curtime
        walltime = curtime - cbinfo.time_start - cbinfo.time_callback

        if cbinfo.task_report:
            memusage = get_memory_usage_kb()
            printlog("\nepoch={:05d}".format(epoch))
            if pinfo and 'norms' in pinfo:
                norms, names = pinfo['norms'], pinfo['names']
                printlog('residual: ' + ', '.join(
                    '{}:{:.5g}'.format(name or str(i), np.array(norm))
                    for i, (norm, name) in enumerate(zip(norms, names))))
            if report_func is not None:
                report_func(problem, state, epoch, cbinfo)
            printlog("memory: {:} MiB".format(memusage // 1024))
            printlog("walltime: {:.3f} s".format(walltime))
            printlog("walltime+callback: {:.3f} s".format(  #
                walltime + cbinfo.time_callback))
            if epoch > cbinfo.epoch:
                wte = (walltime - cbinfo.walltime) / (epoch - cbinfo.epoch)
                printlog("walltime/epoch: {:.3f} ms".format(wte * 1000))
                printlog("throughput: {:.3f}M cells/s".format(
                    np.prod(domain.cshape) / wte / 1e6))
                cbinfo.walltime = walltime
                cbinfo.epoch = epoch

        if cbinfo.task_history:
            memusage = get_memory_usage_kb()
            history.append('epoch', epoch)
            history.append('frame', cbinfo.frame)
            if pinfo and 'norms' in pinfo:
                norms, names = pinfo['norms'], pinfo['names']
                for i, (norm, name) in enumerate(zip(norms, names)):
                    history.append('norm_{:}'.format(name or str(i)),
                                   np.array(norm))
            if pinfo and 'loss' in pinfo:
                history.append('loss', pinfo['loss'])
            if args.linsolver_history and 'linsolver' in pinfo:
                for key, val in pinfo['linsolver'].items():
                    if isinstance(val, (int, float, str, np.floating)):
                        history.append('lin_' + key, val)
            history.append('walltime', walltime)
            history.append('memory', memusage / 1024)
            if history_func is not None:
                history_func(problem, state, epoch, history, cbinfo)
            history.write()

        if cbinfo.task_plot:
            if plot_func is not None:
                plot_func(problem, state, epoch, cbinfo.frame, cbinfo)
            cbinfo.frame += 1

        if cbinfo.task_checkpoint:
            if checkpoint_func is not None:
                checkpoint_func(problem, state, epoch, cbinfo)
            else:
                from .core import checkpoint_save
                path = "checkpoint_{:06d}.pickle".format(epoch)
                printlog(path)
                checkpoint_save(domain, state, path)

        # Subtract second part of callback time.
        curtime = time.time()
        cbinfo.time_callback += time.time() - time_prev
        time_prev = curtime

    callback.cbinfo = cbinfo
    return callback



import numpy as np
from argparse import Namespace


class Domain():

    def __init__(self,
                 domain=None,
                 ndim=None,
                 lower=None,
                 upper=None,
                 dimnames=None,
                 dtype=None,
                 cshape=None):
        domain = domain or Namespace(ndim=None,
                                     lower=0.,
                                     upper=1.,
                                     dimnames=None,
                                     dtype=None,
                                     cshape=None)
        dtype = dtype or domain.dtype
        cshape = cshape or domain.cshape
        ndim = ndim or domain.ndim
        dimnames = dimnames or domain.dimnames or ['x', 'y', 'z'][:ndim]
        lower = lower if lower is not None else domain.lower
        upper = upper if upper is not None else domain.upper
        ndim = len(cshape)
        lower = (np.ones(ndim, dtype=dtype) * lower).astype(dtype)
        upper = (np.ones(ndim, dtype=dtype) * upper).astype(dtype)
        self.ndim = ndim
        self.lower = lower
        self.upper = upper
        self.dimnames = dimnames
        self.dtype = dtype
        self.cshape = cshape

    @staticmethod
    def _names_to_indices(dims, dimnames):
        '''
        Converts a list of direction names to indices.
        Examples with `dimnames = ['t', 'x', 'y']`:
            complete_dims(['x']) -> [1]
            complete_dims(['x', 'y']) -> [1, 2]
            complete_dims([]) -> [0, 1, 2]
        '''
        res = dims if dims is not None and len(dims) else range(len(dimnames))
        res = tuple(
            dimnames.index(i) if isinstance(i, str) else i for i in res)
        return res

    def _points_cell_1d(self, d):
        x = np.linspace(self.lower[d],
                        self.upper[d],
                        self.cshape[d],
                        endpoint=False,
                        dtype=self.dtype)
        x += (x[1] - x[0]) * 0.5
        return x

    def _points_node_1d(self, d):
        x = np.linspace(self.lower[d],
                        self.upper[d],
                        self.cshape[d] + 1,
                        dtype=self.dtype)
        return x

    def _points_1d(self, d, loc):
        if loc == 'c':
            return self._points_cell_1d(d)
        elif loc == 'n':
            return self._points_node_1d(d)
        else:
            raise ValueError("Unknown loc=" + loc)

    def points_1d(self, *dims, loc=None):
        loc = loc or 'c' * self.ndim
        idims = self._names_to_indices(dims, self.dimnames)
        res = [self._points_1d(idim, c) for idim, c in zip(idims, loc)]
        if len(dims) == 1:
            return res[0]
        return res



import numpy as np
import pickle

from .util import assert_equal, printlog
from .backend import ModTensorflow
from . import core_min


class Domain:

    def __init__(
        self,
        cshape,
        dimnames=None,
        lower=0.,
        upper=1.,
        dtype=None,
        multigrid=False,
        mg_convert_all=True,
        mg_nlvl=None,
        mg_factors=None,
        mg_axes=None,
        mg_interp=None,
        mod=None,
    ):
        '''
        cshape: `tuple`
            Grid size measured in cells.
        multigrid: `bool`
            Generate multigrid hierarchy.
        mg_convert_all: `bool`
            Convert all fields to multigrid fields in `init_state()`.
        mg_nlvl: `int`
            Number of multigrid levels. Defaults to maximum possible.
        mg_factors: `list` of `int`
            Factors of each level. Defaults to ones.
        mg_axes: `list` of `bool`
            Axes in which to use multigrid decomposition. Defaults to all.
        mg_interp: `str`
            Multigrid interpolation method. See `interp_to_finer()`.
        '''
        ndim = len(cshape)
        dimnames = dimnames or ['x', 'y', 'z'][:ndim]
        if mod is None:
            from .runtime import mod
        assert_equal(len(dimnames), ndim, f'with dimnames={dimnames}')
        self.ndim = ndim
        self.cshape = cshape
        self.dimnames = dimnames
        if dtype is None:
            from . import runtime
            dtype = runtime.dtype
        self.dtype = dtype
        self.lower = (np.ones(ndim, dtype=dtype) * lower).astype(dtype)
        self.upper = (np.ones(ndim, dtype=dtype) * upper).astype(dtype)
        self.mod = mod

        # Multigrid decomposition
        self.multigrid = multigrid
        if multigrid:
            self.mg_factors = mg_factors
            mg_axes = mg_axes or [True] * ndim
            nlvl_max = min(
                round(np.log2(n)) if ax else max(cshape)
                for n, ax in zip(cshape, mg_axes))
            if mg_nlvl is not None:
                assert mg_nlvl >= 1
                mg_nlvl = min(mg_nlvl, nlvl_max)
            else:
                mg_nlvl = nlvl_max
            self.mg_nlvl = mg_nlvl
            self.mg_cshapes = [
                tuple(n >> lvl if ax else n for n, ax in zip(cshape, mg_axes))
                for lvl in range(mg_nlvl)
            ]
            check_multigrid_cshapes(self.mg_cshapes, mg_axes)
            self.mg_axes = mg_axes
            self.mg_interp = mg_interp
            self.mg_convert_all = mg_convert_all

    @staticmethod
    def _names_to_indices(dims, dimnames):
        '''
        Converts a list of direction names to indices.
        Examples with `dimnames = ['t', 'x', 'y']`:
            complete_dims(['x']) -> [1]
            complete_dims(['x', 'y']) -> [1, 2]
            complete_dims([]) -> [0, 1, 2]
        '''
        res = dims if dims is not None and len(dims) else range(len(dimnames))
        res = tuple(
            dimnames.index(i) if isinstance(i, str) else i for i in res)
        return res

    def cast(self, value, dtype=None):
        dtype = dtype or self.dtype
        return self.mod.cast(value, dtype)

    def get_minimal(self):
        return core_min.Domain(self)

    def _points_cell_1d(self, d):
        x = np.linspace(self.lower[d],
                        self.upper[d],
                        self.cshape[d],
                        endpoint=False,
                        dtype=self.dtype)
        if len(x) > 1:
            x += (x[1] - x[0]) * 0.5
        return x

    def _points_node_1d(self, d):
        x = np.linspace(self.lower[d],
                        self.upper[d],
                        self.cshape[d] + 1,
                        dtype=self.dtype)
        return x

    def _points_1d(self, d, loc):
        if loc == 'c':
            return self._points_cell_1d(d)
        elif loc == 'n':
            return self._points_node_1d(d)
        else:
            raise ValueError("Unknown loc=" + loc)

    def points_1d(self, *dims, loc=None):
        loc = loc or 'c' * self.ndim
        idims = self._names_to_indices(dims, self.dimnames)
        res = [self._points_1d(idim, c) for idim, c in zip(idims, loc)]
        if len(dims) == 1:
            return res[0]
        return res

    def points(self, *dims, loc=None):
        loc = loc or 'c' * self.ndim
        assert_equal(len(loc), self.ndim, f"with loc={loc}")
        dimnames = [v for v, c in zip(self.dimnames, loc) if c != '.']
        idims = self._names_to_indices(dims, dimnames)
        xx = [
            self._points_1d(d, loc[d]) if loc[d] != '.' else None
            for d in range(self.ndim)
        ]
        xx = [x for x in xx if x is not None]
        data = self.mod.meshgrid(*xx, indexing='ij')
        res = tuple(data[i] for i in idims)
        if len(dims) == 1:
            return res[0]
        return res

    def _indices_cell_1d(self, d):
        return np.arange(self.cshape[d], dtype=int)

    def _indices_node_1d(self, d):
        return np.arange(self.cshape[d] + 1, dtype=int)

    def _indices_1d(self, d, loc):
        if loc == 'c':
            return self._indices_cell_1d(d)
        elif loc == 'n':
            return self._indices_node_1d(d)
        else:
            raise ValueError("Unknown loc=" + loc)

    def indices(self, *dims, loc=None):
        loc = loc or 'c' * self.ndim
        dimnames = [v for v, c in zip(self.dimnames, loc) if c in 'cn']
        idims = self._names_to_indices(dims, dimnames)
        xx = [
            self._indices_1d(d, loc[d]) if loc[d] in 'cn' else None
            for d in range(self.ndim)
        ]
        xx = [x for x in xx if x is not None]
        data = self.mod.meshgrid(*xx, indexing='ij')
        res = tuple(data[i] for i in idims)
        if len(dims) == 1:
            return res[0]
        return res

    @staticmethod
    def _get_field_shape(cshape, loc=None):
        '''
        Returns the size of array with given location of value.

        cshape: `tuple`
            Domain shape measured in cells.
        loc: `str`
            Location of value in cell.
            One character per direction: 'c' for center and 'n' for node.
        '''
        loc = loc or 'c' * len(cshape)
        assert all(c in 'cn' for c in loc)
        return tuple(s + 1 if c == 'n' else s for s, c in zip(cshape, loc))

    def get_field_shape(self, loc=None):
        return _get_field_shape(self.cshape, loc=loc)

    def _size_1d(self, d, loc):
        if loc == 'c':
            return self.cshape[d]
        elif loc == 'n':
            return self.cshape[d] + 1
        else:
            raise ValueError("Unknown loc=" + loc)

    def size(self, *dims, loc=None):
        loc = loc or 'c' * self.ndim
        assert_equal(len(loc), self.ndim, f"with loc={loc}")
        idims = self._names_to_indices(dims, self.dimnames)
        res = [self._size_1d(i, loc[i]) for i in idims]
        if len(dims) == 1:
            return res[0]
        return res

    def step_by_dim(self, i):
        return (self.upper[i] - self.lower[i]) / self.cshape[i]

    def step(self, *dims):
        '''
        Returns grid spacing in directions `dims`.
        Examples:
            dx = ctx.step('x')
            dx, = ctx.step() # With dimnames=['x'].
            dx, dy = ctx.step('x', 'y')
            dx, dy = ctx.step() # With dimnames=['x', 'y'].
        '''
        idims = self._names_to_indices(dims, self.dimnames)
        res = tuple(self.step_by_dim(i) for i in idims)
        if len(dims) == 1:
            return res[0]
        return res

    def random_inner(self, size):
        res = latin_hypercube(self.ndim, size, dtype=self.dtype).T
        for i in range(self.ndim):
            res[i] = self.lower[i] + (self.upper[i] - self.lower[i]) * res[i]
        res = [p for p in res]
        return res

    def random_boundary(self, normal, side, size):
        '''
        Returns random points from boundary (domain face).
        normal: `int`
            Direction of normal to boundary, [0, ndim)
        side: `int`
            Side of boundary (0 or 1).
        size: `int`
            Number of samples.
        '''
        assert normal < self.ndim
        assert side == 0 or side == 1
        dtype = self.dtype
        res = latin_hypercube(self.ndim - 1, size, dtype=dtype).T
        const = np.ones(size, dtype=dtype) * side
        res = np.vstack((res[:normal], const, res[normal:]))
        for i in range(self.ndim):
            res[i] = self.lower[i] + (self.upper[i] - self.lower[i]) * res[i]
        res = [p for p in res]
        return res

    def multigrid_to_regular(self, mgfield):
        '''
        Converts multigrid components to regular field.
        uw: `ndarray`
            Multigrid components on levels `self.mg_cshapes`.
        '''
        mod = self.mod
        factors = mgfield.factors or self.mg_factors or [1] * len(cshapes)
        axes = mgfield.axes or self.mg_axes
        shapes = [
            self._get_field_shape(term.cshape, loc=mgfield.loc)
            for term in mgfield.terms
        ]
        assert_equal(len(factors), len(mgfield.terms))
        assert_equal(len(axes), len(mgfield.terms[0].cshape))
        method = mgfield.method or self.mg_interp
        arrays = [term.array * f for term, f in zip(mgfield.terms, factors)]
        loc = ''.join(l if ax else '.' for l, ax in zip(mgfield.loc, axes))
        res = arrays[-1]
        for array in reversed(arrays[:-1]):
            res = array + interp_to_finer(res, loc, method, mod)
        return Field(res, loc=mgfield.loc)

    def get_regular_array(self, field):
        '''
        Returns array from Field, MultigridField, or Array.
        '''
        if isinstance(field, (Field, Array)):
            return field.array
        elif isinstance(field, MultigridField):
            return self.multigrid_to_regular(field).array
        else:
            raise TypeError("Expected Field or MultigridField got {}".format(
                type(field).__name__))

    def regular_to_multigrid(self,
                             field,
                             cshapes=None,
                             factors=None,
                             method=None):
        '''
        Converts regular field to multigrid field.
        u: `ndarray` or `Field`
            Field on the fine grid.
        cshapes: `list` of `tuple`
            List of grid sizes measured in cells, starting from the fine grid.
        '''
        mod = self.mod
        if isinstance(field, (MultigridField, NeuralNet)):
            raise TypeError("Expected Field or ndarray, got type {}".format(
                type(net).__name__))

        field = self.init_field(field)
        cshapes = cshapes or self.mg_cshapes
        factors = factors or self.mg_factors or [1] * len(cshapes)
        assert_equal(len(cshapes), len(factors))
        method = method or self.mg_interp
        terms = [
            Field(field.array / factors[0], loc=field.loc, cshape=field.cshape)
        ]
        for cshape in cshapes[1:]:
            array = mod.zeros(self._get_field_shape(cshape, loc=field.loc),
                              dtype=self.dtype)
            terms.append(Field(array, loc=field.loc, cshape=cshape))
        return MultigridField(terms=terms,
                              loc=field.loc,
                              factors=factors,
                              method=method)

    def init_field(self, field):
        '''
        Initializes field in the given backend.
        Fills omitted attributes with default values.
        '''
        mod = self.mod
        if field is None:
            return self.init_field(
                Field(None, loc='c' * self.ndim, cshape=self.cshape))
        elif isinstance(field, np.ndarray) or mod.is_tensor(field):
            array = field
            return self.init_field(
                Field(array, loc='c' * len(array.shape), cshape=array.shape))
        elif isinstance(field, Field):
            cshape = field.cshape or self.cshape
            ndim = len(cshape)
            loc = field.loc or 'c' * ndim
            assert_equal(len(loc), ndim)
            array = field.array
            if array is None:
                array = mod.zeros(self._get_field_shape(cshape, loc=loc),
                                  dtype=self.dtype)
            array = mod.variable(array, dtype=self.dtype)
            assert_equal(array.shape, self._get_field_shape(cshape, loc=loc))
            return Field(array, loc=loc, cshape=cshape)
        elif isinstance(field, MultigridField):
            return MultigridField(
                [self.init_field(term) for term in field.terms],
                loc=field.loc,
                factors=field.factors,
                axes=field.axes,
                method=field.method)
        elif isinstance(field, NeuralNet):
            dtype = self.dtype
            weights = [mod.variable(w, dtype=dtype) for w in field.weights]
            biases = [mod.variable(b, dtype=dtype) for b in field.biases]
            return NeuralNet(weights,
                             biases,
                             func_in=field.func_in,
                             func_out=field.func_out,
                             activation=field.activation)
        elif isinstance(field, list):
            u = mod.cast(mod.array(field), self.dtype)
            return self.init_field(Array(u, shape=u.shape))
        elif isinstance(field, Array):
            array = field.array
            if array is None:
                array = mod.zeros(field.shape, dtype=self.dtype)
            array = mod.variable(array, dtype=self.dtype)
            return Array(array, field.shape)
        else:
            raise TypeError("Unknown field type '{}'".format(
                type(field).__name__))

    def init_state(self, state):
        '''
        Initializes state variables in the given backend.
        '''
        fields = dict()
        for key in state.fields:
            field = state.fields[key]
            field = self.init_field(field)
            if self.multigrid and self.mg_convert_all and \
                    not isinstance(field, (MultigridField, NeuralNet, Array)):
                field = self.regular_to_multigrid(state.fields[key])
            fields[key] = field
        return State(fields=fields, initialized=True)

    def arrays_from_field(self, field):
        '''
        Returns list of data arrays of `field`.
        '''
        mod = self.mod
        if isinstance(field, Field):
            return [field.array]
        elif isinstance(field, MultigridField):
            return [term.array for term in field.terms]
        elif isinstance(field, NeuralNet):
            return field.weights + field.biases
        elif isinstance(field, Array):
            return [field.array]
        else:
            raise TypeError("Unknown field type '{}'".format(
                type(field).__name__))

    def arrays_from_state(self, state):
        '''
        Returns list of data arrays from fields in `state`.
        '''
        mod = self.mod
        res = []
        for key in state.fields:
            res += self.arrays_from_field(state.fields[key])
        return res

    @staticmethod
    def arrays_to_field(arrays, field):
        '''
        Replaces data in `field` with `arrays`.
        Returns the number of elements in `arrays` consumed.
        '''
        if isinstance(field, Field):
            field.array = arrays[0]
            return 1
        elif isinstance(field, MultigridField):
            offset = 0
            for term in field.terms:
                term.array = arrays[offset]
                offset += 1
            return offset
        elif isinstance(field, NeuralNet):
            offset = 0
            for i in range(len(field.weights)):
                field.weights[i] = arrays[offset]
                offset += 1
            for i in range(len(field.biases)):
                field.biases[i] = arrays[offset]
                offset += 1
            return offset
        elif isinstance(field, Array):
            field.array = arrays[0]
            return 1
        else:
            raise TypeError("Unknown field type '{}'".format(
                type(field).__name__))

    @staticmethod
    def arrays_to_state(arrays, state):
        '''
        Replaces data in `state` with `arrays`.

        Returns the number of elements in `arrays` consumed.
        '''
        offset = 0
        for key in state.fields:
            offset += Domain.arrays_to_field(arrays[offset:],
                                             state.fields[key])
        return offset

    def pack_field(self, field):
        '''
        Returns a flat array with data from `field`.
        '''
        mod = self.mod
        res = self.arrays_from_field(field)
        res = [mod.flatten(f) for f in res]
        return mod.concatenate(res, axis=0)

    def pack_state(self, state):
        '''
        Returns a flat array with data from `state`.
        '''
        mod = self.mod
        res = self.arrays_from_state(state)
        res = [mod.flatten(f) for f in res]
        return mod.concatenate(res, axis=0)

    def unpack_field(self, packed, field):
        '''
        Unpacks a flat array `packed` into `field`.
        Returns the number of elements in `packed` consumed.
        '''
        mod = self.mod
        arrays = self.arrays_from_field(field)
        sizes = [np.prod(a.shape) for a in arrays]
        split = mod.split_by_sizes(packed[:sum(sizes)], sizes)
        arrays = [mod.reshape(s, a.shape) for s, a in zip(split, arrays)]
        self.arrays_to_field(arrays, field)
        return sum(sizes)

    def unpack_state(self, packed, state):
        '''
        Unpacks a flat array `packed` into `state`.
        Returns the number of elements in `packed` consumed.
        '''
        mod = self.mod
        arrays = self.arrays_from_state(state)
        sizes = [np.prod(a.shape) for a in arrays]
        split = mod.split_by_sizes(packed[:sum(sizes)], sizes)
        arrays = [mod.reshape(s, a.shape) for s, a in zip(split, arrays)]
        self.arrays_to_state(arrays, state)
        return sum(sizes)

    def make_neural_net(self,
                        layers,
                        initializer='lecun',
                        func_in=None,
                        func_out=None,
                        activation=None):
        return make_neural_net(layers, self.dtype, self.mod, initializer,
                               func_in, func_out, activation)

    def field(self, state, key, *shift):
        mod = self.mod
        field = state.fields[key]
        if not isinstance(field, (Field, MultigridField, Array)):
            raise TypeError(
                "Expected Field or MultigridField, got type {} for field '{}'".
                format(type(net).__name__, key))
        if isinstance(field, Array):
            if len(shift):
                raise RuntimeError('Array requires an empty shift')
            return field.array
        shift = shift or (0, ) * self.ndim
        if len(shift) != self.ndim:
            raise RuntimeError(
                "Expected {} shift components, got shift={}".format(
                    self.ndim, shift))
        array = self.get_regular_array(field)
        return mod.roll(array, np.negative(shift), range(self.ndim))

    def neural_net(self, state, key):
        net = state.fields[key]
        if not isinstance(net, NeuralNet):
            raise TypeError(
                "Expected NeuralNet, got type {} for key='{}'".format(
                    type(net).__name__, key))
        res = lambda *inputs: eval_neural_net(net, inputs, self.mod)
        return res

    def get_context(self, state, extra=None, tracers=None):
        ctx = Context(self.domain,
                      state,
                      watch_func=lambda _: None,
                      extra=extra,
                      tracers=tracers,
                      distinct_shift=False)
        return ctx


class Field:

    def __init__(self, array=None, loc=None, cshape=None):
        '''
        array: `ndarray`
            Data array.
        loc: `str`
            Location in cell.
            One character per direction: 'c' for center and 'n' for node.
        cshape: `tuple`
            Grid size measured in cells.
        '''
        self.array = array
        self.loc = loc
        self.cshape = cshape

    def __str__(self):
        return "odil.Field({}, loc={}, cshape={:}>".format(
            str(self.array), self.loc, self.cshape)

    def __repr__(self):
        return "odil.Field({}, loc='{}', cshape={:})".format(
            repr(self.array), self.loc, self.cshape)


class MultigridField:

    def __init__(self,
                 terms=None,
                 loc=None,
                 factors=None,
                 axes=None,
                 method=None):
        '''
        terms: `list` of `Field`
            Terms of the multigrid decomposition.
        loc: `str`
            Location in cell.
            One character per direction: 'c' for center and 'n' for node.
        axes: `list` of `bool`
            Flag for each axis. If False, the axis is unchanged.
        factors: `list` of `float`
            Factor of each term in the decomposition. Defaults to 1.
        method: `str`
            Interpolation method.
        '''
        self.terms = terms
        self.loc = loc
        self.factors = factors
        self.axes = axes
        self.method = method


class NeuralNet:

    def __init__(self,
                 weights=None,
                 biases=None,
                 func_in=None,
                 func_out=None,
                 activation=None):
        '''
        weights: `list` of `ndarray`
            List of weights, matrices of size `(no, ni)` for each
            layer with `ni` inputs and `no` outputs.
        biases: `list` of `ndarray`
            List of biases,  arrays. of size `no` for each
            layer with `no` outputs.
        func_in: `callable`
            Function called on inputs.
        func_out: `callable`
            Function called on outputs.
        activation: `str`
            Name of the activation function.
        '''
        self.weights = weights
        self.biases = biases
        self.func_in = func_in
        self.func_out = func_out
        self.activation = activation or 'tanh'


class Array:

    def __init__(self, array=None, shape=None):
        '''
        array: `ndarray`
            Data array.
        shape: `tuple`
            Array shape.
        '''
        self.array = array
        self.shape = shape

    def __str__(self):
        return "odil.Array({}, shape={:}>".format(str(self.array), self.shape)

    def __repr__(self):
        return "odil.Array({}, shape={:})".format(repr(self.array), self.shape)


class State:

    def __init__(self, fields=None, initialized=False):
        '''
        fields: A dict mapping field names to state variables
            of type Field, MultigridField, or NeuralNet.
        '''
        self.fields = fields if fields is not None else dict()
        self.initialized = initialized


def interp_to_finer(u, loc=None, method=None, mod=None, depth=1):
    '''
    Interpolates a field to a finer grid.

    u: `array`
        Input field.
    loc: `str`
        Location of value in cell, one character per direction.
        'c': cell, new size `2 * n`;
        'n': node, new size `2 * (n - 1) + 1`;
        '.': none, new size `n`.
    method: `str`
        Interpolation method.
        'conv': using the transpose of convolution,
                may be slower than 'stack' with JIT (JAX and TF XLA).
        'stack': using a stack of shifted arrays.
    depth: `int`
        Number of repetitions.
    '''
    if depth == 0:
        return u

    method = method or 'stack'
    if method not in ['conv', 'stack']:
        raise ValueError("Unknown method='{}'".format(method))
    assert_equal(len(loc), len(u.shape))
    for l in loc:
        assert l in 'cn.', "Invalid loc={}".format(loc)
    dim = len(u.shape)

    # Add padding depending on value location:
    # 'c': linear extrapolation,
    # 'n': no padding,
    # '.': no padding.
    pad_width = [(1, 1) if l == 'c' else (0, 0) for l in loc]
    ur = mod.pad(u, pad_width=pad_width, mode='reflect')
    us = mod.pad(u, pad_width=pad_width, mode='symmetric')
    upad = 2 * us - ur

    if method == 'conv':
        # Convolution weights.
        wnode = np.array([1, 2, 1]) * 0.5
        wcell = np.array([1, 3, 3, 1]) * 0.25
        wnone = np.array([1.])
        wloc = {'n': wnode, 'c': wcell, '.': wnone}
        w = wloc[loc[0]]
        for i in range(1, dim):
            w = np.kron(wloc[loc[i]], w[..., None])
        w = mod.cast(mod.reshape(w, w.shape + (1, 1)), u.dtype)

        # Output shape.
        oshape = lambda s: {'n': s * 2 + 1, 'c': s * 2 + 2, '.': s}
        oshape = tuple(oshape(s)[l] for l, s in zip(loc, upad.shape))
        oshape = (1, ) + oshape + (1, )
        strides = tuple(1 if l == '.' else 2 for l in loc)
        upad = mod.reshape(upad, (1, ) + upad.shape + (1, ))
        res = mod.conv_transpose(upad,
                                 filters=w,
                                 output_shape=oshape,
                                 strides=strides,
                                 padding='VALID')
        # Remove edges.
        oslice = {'n': slice(1, -1), 'c': slice(3, -3), '.': slice(0, None)}
        res = res[(0, ) + tuple(oslice[l] for l in loc) + (0, )]
    elif method == 'stack':

        def term(*dd, ww=None):
            dd = [tuple(-v for v in d) for d in dd]
            return sum(w * mod.roll(upad, d, range(dim))
                       for d, w in zip(dd, ww) if w) / sum(ww)

        # Offsets of nodes of a dim-dimensional cube.
        dd = np.meshgrid(*[[0] if l == '.' else [0, 1] for l in loc],
                         indexing='ij')
        dshape = tuple(1 if l == '.' else 2 for l in loc)
        # Example dim=2: dd = [(0,0), (0,1), (1,0), (1,1)].
        dd = np.reshape(dd, (dim, -1)).T
        # Indices with location in node.
        sn = [i for i, l in enumerate(loc) if l == 'n']
        # Indices with location in cell.
        sc = [i for i, l in enumerate(loc) if l == 'c']

        def weight(r, d):
            return (3**(sum(1 - abs(r - d)[sc]))  #
                    if np.all((r - d)[sn] <= 0) else 0)

        uu = [term(*dd, ww=[weight(r, d) for r in dd]) for d in dd]
        res = mod.stack(uu)
        res = mod.reshape(res, dshape + upad.shape)
        for i in range(dim):
            res = [res[i] for i in range(res.shape[0])]
            res = mod.stack(res, axis=dim + i)
        res = mod.reshape(res,
                          tuple(s * d for s, d in zip(upad.shape, dshape)))
        # Remove edges.
        oslice = {'n': slice(0, -1), 'c': slice(1, -3), '.': slice(0, None)}
        res = res[tuple(oslice[l] for l in loc)]
    else:
        raise ValueError('Unknown method=' + method)

    return interp_to_finer(res, loc, method, mod, depth - 1)


def restrict_to_coarser(u, loc=None, method=None, mod=None, depth=1):
    '''
    Restricts a field to a coarser grid.

    u: `array`
        Input field.
    loc: `str`
        Location of value in cell, one character per direction.
        'c': cell, new size `n // 2`;
        'n': node, new size `(n - 1) // 2 + 1`;
        '.': none, new size `n`.
    method: `str`
        Restriction method.
        'conv': using the convolution
    depth: `int`
        Number of repetitions.
    '''
    if depth == 0:
        return u

    method = method or 'conv'
    if method not in ['conv']:
        raise ValueError("Unknown method='{}'".format(method))
    assert_equal(len(loc), len(u.shape))
    for l in loc:
        assert l in 'cn.', "Invalid loc={}".format(loc)
    dim = len(u.shape)

    # Add padding depending on value location:
    # 'c': no padding,
    # 'n': linear extrapolation, combined with the [1,2,1] kernel
    #      implements the identity condition on the boundaries
    # '.': no padding.
    pad_width = [(1, 1) if l == 'n' else (0, 0) for l in loc]
    ur = mod.pad(u, pad_width=pad_width, mode='reflect')
    us = mod.pad(u, pad_width=pad_width, mode='symmetric')
    upad = 2 * us - ur

    if method == 'conv':
        # Convolution weights.
        wnode = np.array([1, 2, 1]) * 0.25
        wcell = np.array([1, 1]) * 0.5
        wnone = np.array([1.])
        wloc = {'n': wnode, 'c': wcell, '.': wnone}
        w = wloc[loc[0]]
        for i in range(1, dim):
            w = np.kron(wloc[loc[i]], w[..., None])
        w = mod.cast(w, u.dtype)
        res = mod.convolution(upad, filters=w, strides=2, padding='VALID')
    else:
        raise ValueError('Unknown method=' + method)

    return restrict_to_coarser(res, loc, method, mod, depth - 1)


def check_multigrid_cshapes(cshapes, axes=None):
    '''
    Checks grid sizes in a multigrid hierarchy.

    cshapes: `list` of `tuple`
        List of grid sizes measured in cells, starting from the fine grid.
    '''
    if not len(cshapes):
        return
    dim = len(cshapes[0])
    axes = axes or [True] * dim
    assert_equal(len(axes), dim)
    for i in range(1, len(cshapes)):
        for j in range(dim):
            if axes[j]:
                n = cshapes[i][j]
                nfine_expect = n * 2
                nfine = cshapes[i - 1][j]
                assert_equal(nfine, nfine_expect,
                             ' with cshapes={:}'.format(cshapes))


def make_neural_net(layers,
                    dtype,
                    mod,
                    initializer='lecun',
                    func_in=None,
                    func_out=None,
                    activation=None):
    '''
    Returns a NeuralNet with random initial weights and zero biases.

    layers: `list` of `int`
        Number of neuron in each layer.
    '''

    def get_scale(ni, no):
        if initializer == 'legacy':
            return np.sqrt(1. / ni)
        elif initializer == 'glorot':
            return np.sqrt(6. / (ni + no))
        elif initializer == 'lecun':
            return np.sqrt(3. / ni)
        elif initializer == 'he':
            return np.sqrt(6. / ni)
        raise ValueError('Unknown initializer=' + initializer)

    weights = []
    biases = []
    for ni, no in zip(layers[:-1], layers[1:]):
        scale = get_scale(ni, no)
        weights.append(  #
            mod.random.uniform(shape=(no, ni),
                               minval=-scale,
                               maxval=scale,
                               dtype=dtype))
        biases.append(mod.zeros(no, dtype=dtype))
    return NeuralNet(weights,
                     biases,
                     func_in=func_in,
                     func_out=func_out,
                     activation=activation)


def eval_neural_net(net: NeuralNet, inputs, mod, frozen=False):
    '''
    Evaluates a fully connected neural network.

    net: `NeuralNet`
        Neural network to evaluate.
    inputs: `list` of `ndarray`
        List of input arrays. All arrays must have the same shape.
    frozen: `bool`
        Apply `stop_gradient()` to the weights.

    Returns: `list` of `ndarray`
        List of output arrays.
        The output arrays have the same shape as the input arrays.
    '''
    weights = net.weights
    biases = net.biases

    # Check shapes of weights and biases.
    assert_equal(len(weights), len(biases), 'Weights and biases do not match')
    assert_equal(weights[0].shape[1], len(inputs),
                 'Weights and inputs do not match')
    for w, b in zip(weights, biases):
        assert_equal(w.shape[0], b.shape[0])

    if frozen:
        weights = [mod.stop_gradient(w) for w in weights]
        biases = [mod.stop_gradient(b) for b in biases]

    func_act = {
        'tanh': mod.tanh,
        'relu': mod.relu,
        'none': lambda x: x,
    }[net.activation]

    n = len(weights)
    if net.func_in is not None:
        inputs = net.func_in(*inputs)
    tmp = mod.stack(inputs, axis=0)
    axes = range(len(tmp.shape))
    perm = np.roll(axes, -1)  # (1, ..., N-1, 0)
    # Let the last two dimensions of `tmp` be columns of inputs.
    tmp = mod.transpose(tmp, perm)[..., None]
    for i in range(n):
        w = weights[i]
        b = biases[i]
        # Multiply matrix `w` by each column of `tmp` and add bias.
        tmp = mod.matmul(w, tmp) + b[:, None]
        if i < n - 1:
            tmp = func_act(tmp)
    perm = np.roll(axes, 1)  # (N-1, 0, ..., N-2)
    # Let the last two dimensions of `tmp` be columns of inputs.
    tmp = mod.transpose(tmp[..., 0], perm)
    outputs = [tmp[i] for i in range(tmp.shape[0])]
    if net.func_out is not None:
        outputs = net.func_out(*outputs)
    return outputs


class Context:

    class Raw:

        def __init__(self, value):
            self.value = value

    def __init__(self,
                 domain,
                 state,
                 watch_func=None,
                 extra=None,
                 tracers=None,
                 distinct_shift=False):
        '''
        watch_func: `callable`
            Callback to add an array to watched variables.
        distinct_shift: `bool`
            If True, fields with different shifts returned by `field()` are
            considered distinct symbols, which is used by `linearize()`
            and `eval_operator_grad()`.
        extra:
            Regular Python value to be available as `ctx.extra`.
            Typically an `argparse.Namespace` object.
            Changing this value does not trigger recompilation.
        tracers: `dict`
            Dictionary of arrays, scalars, or standard Python containers thereof.
            This argument becomes part of the signature of the jitted function.
        '''
        self.domain = domain
        self.state = state
        self.watch_func = watch_func or (lambda _: None)
        self.extra = extra
        self.tracers = tracers
        self.dtype = domain.dtype
        self.mod = domain.mod
        self.distinct_shift = distinct_shift
        # Storage for field arguments. Mapping from (key, shift, loc) to array.
        self.desc_to_array = dict()
        # Storage for field arguments to request full Jacobian.
        self.key_to_array_jac = dict()
        # Aliases for Domain methods.
        self.step = domain.step
        self.size = domain.size
        self.indices = domain.indices
        self.points = domain.points

    def cast(self, value, dtype=None):
        dtype = dtype or self.dtype
        return self.mod.cast(value, dtype)

    def field(self, key, *shift, loc=None, frozen=False):
        domain = self.domain
        mod = domain.mod
        field = self.state.fields[key]
        if not isinstance(field, (Field, MultigridField, Array)):
            raise TypeError(
                "Expected Field or MultigridField, got type {} for key='{}'".
                format(type(net).__name__, key))
        if isinstance(field, Array):
            if len(shift):
                raise RuntimeError('Array requires an empty shift')
            self.watch_func(field.array)
            self.key_to_array_jac[(key, None, None)] = field.array
            array = field.array
            if frozen:
                array = mod.stop_gradient(array)
            return array
        shift_src = (0, ) * domain.ndim
        shift = shift or shift_src
        loc = loc or field.loc
        if len(shift) != domain.ndim:
            raise RuntimeError(
                "Expected {} shift components, got shift={}".format(
                    domain.ndim, shift))
        desc = (key, shift, loc)  # Descriptor of target field.
        desc_src = (key, shift_src, field.loc)  # Descriptor of source field.
        if desc in self.desc_to_array:  # Reuse target field.
            array = self.desc_to_array[desc]
            if self.distinct_shift and isinstance(field, Field):
                self.watch_func(array)
        else:
            if desc_src in self.desc_to_array:  # Reuse source field.
                array_src = self.desc_to_array[desc_src]
            else:  # Create new source field.
                if not self.distinct_shift:
                    if isinstance(field, Field):
                        self.watch_func(field.array)
                    elif isinstance(field, MultigridField):
                        for term in field.terms:
                            self.watch_func(term.array)
                array_src = domain.get_regular_array(field)
                self.desc_to_array[desc_src] = array_src
            if self.distinct_shift and desc != desc_src:
                # If shifts are treated separately,
                # avoid differentiation through the source field.
                array_src = mod.stop_gradient(array_src)
            array = array_src
            # True if needs padding in that direction.
            pad_flag = [
                lf == 'c' and l == 'n' for lf, l in zip(field.loc, loc)
            ]
            # Pad the array to account for change of location.
            if any(pad_flag):
                pad_width = [(1, 0) if f else (0, 0) for f in pad_flag]
                array = mod.pad(array, pad_width=pad_width, mode='constant')
            # Shift the array.
            if shift != shift_src:
                array = mod.roll(array, np.negative(shift), range(domain.ndim))
            # True if needs trimming in that direction.
            trim_flag = [
                lf == 'n' and l == 'c' for lf, l in zip(field.loc, loc)
            ]
            # Trim the array to account for change of location.
            if any(trim_flag):
                trim_slice = [slice(0, -1 if f else None) for f in trim_flag]
                array = array[trim_slice]
            if self.distinct_shift and isinstance(field, Field):
                self.watch_func(array)
            self.desc_to_array[desc] = array
        if frozen:
            array = mod.stop_gradient(array)
        return array

    def neural_net(self, key, frozen=False):
        domain = self.domain
        net = self.state.fields[key]
        if not isinstance(net, NeuralNet):
            raise TypeError(
                "Expected NeuralNet, got type {} for key='{}'".format(
                    type(net).__name__, key))
        arrays = domain.arrays_from_field(net)
        self.watch_func(arrays)
        if self.distinct_shift:
            self.key_to_array_jac[(key, None, None)] = arrays
        res = lambda *inputs: eval_neural_net(
            net, inputs, self.mod, frozen=frozen)
        return res


class Problem:

    def __init__(self, operator, domain, extra=None, tracers=None, jit=None):
        '''
        operator: callable(mod, ctx)
            Discrete operator returning fields on grid.
            Each field corresponds to an equation to be solved.
            Arguments are:
                mod: module with mathematical functions (tensorflow, numpy)
                ctx: instance of Context
        domain: instance of Domain
        '''
        self.domain = domain
        self.operator = operator
        self.extra = extra
        if tracers is None:
            tracers = dict()
        if 'epoch' not in tracers:
            tracers['epoch'] = 0
        self.tracers = tracers
        mod = domain.mod
        if jit is None:
            from .runtime import jit
        self.jit = jit

        self._cache_eval_loss_grad = dict()
        self._cache_eval_operator = dict()
        self._cache_eval_operator_grad = dict()

        if isinstance(mod, ModTensorflow):
            self._eval_loss_grad = self._eval_loss_grad_tf
            self._eval_operator = self._eval_operator_tf
            self._eval_operator_grad = self._eval_operator_grad_tf
        elif mod.jax is not None:
            self._eval_loss_grad = self._eval_loss_grad_jax
            self._eval_operator = self._eval_operator_jax
            self._eval_operator_grad = self._eval_operator_grad_jax
        else:
            raise NotImplementedError('Unsupported mod={:}'.format(mod))

    def _eval_loss_grad_tf(self, state):
        domain = self.domain
        mod = domain.mod
        tf = mod.tf
        cache = self._cache_eval_loss_grad

        if 'state' not in cache:
            cache['state'] = domain.init_state(state)

        def func(arrays):
            with tf.GradientTape(persistent=False,
                                 watch_accessed_variables=False) as tape:
                domain.arrays_to_state(arrays, cache['state'])
                ctx = Context(domain,
                              cache['state'],
                              watch_func=tape.watch,
                              extra=self.extra,
                              tracers=self.tracers)
                ff = self.operator(ctx)
                assert isinstance(ff, (tuple, list)) and len(ff), \
                    'Operator must return a non-empty list'
                names = [f[0] if isinstance(f, tuple) else '' for f in ff]
                values = [f[1] if isinstance(f, tuple) else f for f in ff]
                terms = [
                    mod.mean(v.value) if isinstance(v, Context.Raw)  #
                    else mod.mean(mod.square(v)) for v in values
                ]
                loss = sum(terms)
                norms = [
                    t if isinstance(v, Context.Raw)  #
                    else mod.sqrt(t) for t, v in zip(terms, values)
                ]
            grads = tape.gradient(loss, arrays)
            grads = [
                g if g is not None else mod.zeros_like(u)
                for u, g in zip(arrays, grads)
            ]
            if 'names' not in cache:
                cache['names'] = names
            return loss, grads, terms, norms

        if 'func' not in cache:
            cache['func'] = tf.function(func, jit_compile=self.jit)

        # Evaluate gradients.
        arrays = domain.arrays_from_state(state)
        loss, grads, terms, norms = cache['func'](arrays)
        return loss, grads, terms, cache['names'], norms

    def _eval_loss_grad_jax(self, state):
        domain = self.domain
        mod = domain.mod
        jax = mod.jax
        cache = self._cache_eval_loss_grad

        def eval_loss(arrays, tracers):
            if 'state' not in cache:
                cache['state'] = domain.init_state(state)
            domain.arrays_to_state(arrays, cache['state'])
            ctx = Context(domain,
                          cache['state'],
                          extra=self.extra,
                          tracers=tracers)
            ff = self.operator(ctx)
            assert isinstance(ff, (tuple, list)) and len(ff), \
                'Operator must return a non-empty list'
            names = [f[0] if isinstance(f, tuple) else '' for f in ff]
            values = [f[1] if isinstance(f, tuple) else f for f in ff]
            terms = [
                mod.mean(f.value) if isinstance(f, Context.Raw) \
                else mod.mean(mod.square(f)) for f in values
            ]
            loss = sum(terms)
            norms = [
                t if isinstance(v, Context.Raw)  #
                else mod.sqrt(t) for t, v in zip(terms, values)
            ]
            return loss, (terms, names, norms)

        def func(arrays, tracers):
            # Create gradient function.
            fgrad = jax.value_and_grad(eval_loss, argnums=[0], has_aux=True)
            (loss, (terms, names, norms)), grads = fgrad(arrays, tracers)
            # Use cache to store the names as JAX does not support `str`.
            cache['names'] = names
            return loss, grads[0], terms, norms

        if 'func' not in cache:
            cache['func'] = jax.jit(func)

        arrays = domain.arrays_from_state(state)
        loss, grads, terms, norms = cache['func'](arrays, self.tracers)
        return loss, grads, terms, cache['names'], norms

    def linearize(self, state, modsp=None):
        '''
        Returns a vector V0 and a sparse matrix M that linearize the operator
            operator(V) ~= M @ (V - V0) + V0
        where V is a flattened state.
        '''
        if not state.initialized:
            raise RuntimeError(
                'Uninitialized state, use `state = domain.init_state(state)`')
        domain = self.domain
        mod = domain.mod
        modsp = modsp or mod.modsp
        if modsp is None:
            import scipy
            import scipy.sparse as modsp

        values, grads, names = self.eval_operator_grad(state)

        # Mapping from field key to offset in flattened state.
        key_to_offset = dict()
        # Mapping from field key to size of flattened data.
        key_to_size = dict()
        offset = 0
        # Iterate over unknown fields.
        for key, field in state.fields.items():
            arrays = domain.arrays_from_field(field)
            size = sum(np.prod(array.shape) for array in arrays)
            key_to_offset[key] = offset
            key_to_size[key] = size
            offset += size
        size_all = offset  # Size of the vector of unknowns.
        del offset

        def field_to_matrix(key, shift, loc, field, garray):
            offset = key_to_offset[key]
            size = key_to_size[key]
            cols = offset + mod.reshape(mod.arange(size), field.array.shape)
            # True if needs padding in that direction.
            pad_flag = [
                lf == 'c' and l == 'n' for lf, l in zip(field.loc, loc)
            ]
            # Pad the array to account for change of location.
            if any(pad_flag):
                pad_width = [(1, 0) if f else (0, 0) for f in pad_flag]
                cols = mod.pad(cols, pad_width=pad_width, mode='constant')
            # Shift the array.
            if shift != shift_src:
                cols = mod.roll(cols, np.negative(shift), range(domain.ndim))
            # True if needs trimming in that direction.
            trim_flag = [
                lf == 'n' and l == 'c' for lf, l in zip(field.loc, loc)
            ]
            # Trim the array to account for change of location.
            if any(trim_flag):
                trim_slice = [slice(0, -1 if f else None) for f in trim_flag]
                cols = cols[trim_slice]
            # Indices of rows, i.e. components of operator value.
            rows = mod.arange(np.prod(value.shape))
            # Indices of columns, i.e. components of unknown field.
            cols = mod.flatten(cols)
            garray = mod.flatten(garray)
            matr = modsp.csr_array((garray, (rows, cols)),
                                   dtype=domain.dtype,
                                   shape=(np.prod(value.shape), size_all))
            return matr

        shift_src = (0, ) * domain.ndim
        matrices = []
        vectors = []
        for name, value, grad in zip(names, values, grads):
            # Shape of the resulting matrix.
            mshape = (np.prod(value.shape), size_all)
            # Rows of the resulting matrix corresponding to one operator value.
            matrix = modsp.csr_array(mshape, dtype=domain.dtype)
            for desc, garray in grad.items():
                key, shift, loc = desc
                if garray is None:
                    # Skip empty gradient.
                    continue
                if isinstance(garray, list) and all(a is None for a in garray):
                    # Skip array of empty gradients.
                    continue
                field = state.fields[key]
                if shift is None:  # Full Jacobian.
                    if isinstance(garray, list):
                        garray = [
                            mod.reshape(a, [mshape[0], -1]) for a in garray
                        ]
                        garray = mod.concatenate(garray, axis=1)
                    # Dense array to sparse.
                    m = modsp.csr_array(garray)
                    # Shift the columns to the field's offset.
                    m = modsp.csr_array(
                        (m.data, m.indices + key_to_offset[key], m.indptr),
                        shape=mshape)
                    matrix += m
                else:  # Element-wise gradient.
                    if not isinstance(field, Field):
                        raise TypeError(
                            "Expected Field, got type {} for key='{}'".format(
                                type(field).__name__, key))
                    matrix += field_to_matrix(key, shift, loc, field, garray)
            matrices.append(matrix)
            vectors.append(mod.flatten(value))
        # Append gradients from weights to gradients from fields.
        matrix = modsp.vstack(matrices)
        vector = mod.concatenate(vectors, axis=0)

        return vector, matrix

    def eval_loss_grad(self, state):
        '''
        Evaluates the loss function and its gradient on `state`.

        Returns:
        loss: `float`
            Value of the loss function.
        grads: `list` of `ndarray`
            Gradient of the loss with respect to each array in the state.
        terms: `list` of `float`
            Terms of the loss function (mean squared operator values).
        names: `list` of `str`
            Names of the operator values.
        norms: `list` of `float`
            Norms (root-mean-square) of the operator values.
        '''
        if not state.initialized:
            raise RuntimeError(
                'Uninitialized state, use `state = domain.init_state(state)`')
        loss, grads, terms, names, norms = self._eval_loss_grad(state)
        loss = np.array(loss)
        terms = list(map(np.array, terms))
        norms = list(map(np.array, norms))
        return loss, grads, terms, names, norms

    def _eval_operator_tf(self, state):
        domain = self.domain
        mod = domain.mod
        tf = mod.tf
        cache = self._cache_eval_operator

        if 'state' not in cache:
            cache['state'] = domain.init_state(state)

        def func(arrays, tracers):
            domain.arrays_to_state(arrays, cache['state'])
            ctx = Context(domain,
                          cache['state'],
                          extra=self.extra,
                          tracers=tracers)
            ff = self.operator(ctx)
            assert isinstance(ff, (tuple, list)) and len(ff), \
                'Operator must return a non-empty list'
            names = [f[0] if isinstance(f, tuple) else '' for f in ff]
            values = [f[1] if isinstance(f, tuple) else f for f in ff]
            if 'names' not in cache:
                cache['names'] = names
            return values

        if 'func' not in cache:
            cache['func'] = tf.function(func, jit_compile=self.jit)

        # Evaluate gradients.
        arrays = domain.arrays_from_state(state)
        values = cache['func'](arrays, self.tracers)
        return values, cache['names']

    def _eval_operator_jax(self, state):
        domain = self.domain
        mod = domain.mod
        jax = mod.jax
        cache = self._cache_eval_operator

        def func(arrays, tracers):
            if 'state' not in cache:
                cache['state'] = domain.init_state(state)
            domain.arrays_to_state(arrays, cache['state'])
            ctx = Context(domain,
                          cache['state'],
                          extra=self.extra,
                          tracers=tracers)
            ff = self.operator(ctx)
            assert isinstance(ff, (tuple, list)) and len(ff), \
                'Operator must return a non-empty list'
            names = [f[0] if isinstance(f, tuple) else '' for f in ff]
            values = [f[1] if isinstance(f, tuple) else f for f in ff]
            if 'names' not in cache:
                cache['names'] = names
            return values

        if 'func' not in cache:
            cache['func'] = jax.jit(func)

        arrays = domain.arrays_from_state(state)
        values = cache['func'](arrays, self.tracers)
        return values, cache['names']

    def eval_operator(self, state):
        '''
        Evaluates the operator on `state`.

        Returns:
        values: `list` of `ndarray`
            Arrays containing the operator values.
        names: `list` of `str`
            Names of the operator values.
        '''
        if not state.initialized:
            raise RuntimeError(
                'Uninitialized state, use `state = domain.init_state(state)`')
        values, names = self._eval_operator(state)
        return values, names

    def _eval_operator_grad_tf(self, state):
        domain = self.domain
        mod = domain.mod
        tf = mod.tf
        cache = self._cache_eval_operator_grad

        if 'state' not in cache:
            cache['state'] = domain.init_state(state)

        def func(arrays, tracers):
            with tf.GradientTape(persistent=True,
                                 watch_accessed_variables=False) as tape:

                domain.arrays_to_state(arrays, cache['state'])
                ctx = Context(domain,
                              cache['state'],
                              watch_func=tape.watch,
                              extra=self.extra,
                              tracers=tracers,
                              distinct_shift=True)
                ff = self.operator(ctx)
                assert isinstance(ff, (tuple, list)) and len(ff), \
                    'Operator must return a non-empty list'
                names = [f[0] if isinstance(f, tuple) else '' for f in ff]
                values = [f[1] if isinstance(f, tuple) else f for f in ff]
                sums = [tf.reduce_sum(v) for v in values]
            grads = [tape.gradient(s, ctx.desc_to_array) for s in sums]
            # Having experimental_use_pfor=True leads to excessive memory usage,
            # sufficient to store a dense matrix of size `prod(domain.shape)**2`.
            grads = []
            for v, s in zip(values, sums):
                g = tape.gradient(s, ctx.desc_to_array)
                if ctx.key_to_array_jac:
                    jac = tape.jacobian(v,
                                        ctx.key_to_array_jac,
                                        experimental_use_pfor=False)
                    g.update(jac)
                grads.append(g)
            if 'names' not in cache:
                cache['names'] = names
            return values, grads

        if 'func' not in cache:
            cache['func'] = tf.function(func, jit_compile=self.jit)

        # Evaluate gradients.
        arrays = domain.arrays_from_state(state)
        values, grads = cache['func'](arrays, self.tracers)
        return values, grads, cache['names']

    def _eval_operator_grad_jax(self, state):
        raise NotImplementedError()

    def eval_operator_grad(self, state):
        '''
        Evaluates the operator and its gradient with respect to fields in state.

        Returns:
        values: `list` of `ndarray`
            Arrays containing the operator values.
        grads: `list` of `dict`
            For each operator value, contains a mapping with the gradient
            of that value with respect to stencil fields.
            The mapping is from tuples (key, shift, loc) to array.
        names: `list` of `str`
            Names of the operator values.
        '''
        if not state.initialized:
            raise RuntimeError(
                'Uninitialized state, use `state = domain.init_state(state)`')
        values, grads, names = self._eval_operator_grad(state)
        return values, grads, names

    def get_context(self, state):
        return self.domain.get_context(extra=self.extra, tracers=self.tracers)


def checkpoint_save(domain, state, path):
    '''
    Saves state to a checkpoint file.
    The checkpoint file will contain a mapping with one entry `fields`
    which is itself a mapping from field keys to lists of arrays
    returned by `domain.arrays_from_field()`.

    path: `str`
        Path to the new checkpoint `.pickle` file.
    '''
    res = dict()
    fields = dict()
    for key in state.fields:
        arrays = domain.arrays_from_field(state.fields[key])
        fields[key] = list(map(np.array, arrays))
    res['fields'] = fields

    with open(path, 'wb') as f:
        pickle.dump(res, f)


def checkpoint_load(domain, state, path, skip_missing=True, keys=None):
    '''
    Loads fields from a checkpoint file to state.
    The checkpoint file must contain a mapping with one entry `fields`
    which is itself a mapping from field keys to lists of arrays
    returned by `domain.arrays_from_field()`.

    state: `State`
        A state with defined fields to load.
    skip_missing: `bool`
        If True, ignore fields from `field_to_load` missing in the checkpoint.
    keys: `list` of `str`
        List of field keys to load.
    '''
    with open(path, 'rb') as f:
        s = pickle.load(f)
    data = s.get('fields', dict())
    keys = keys or state.fields.keys()
    for key in keys:
        if key not in data:
            if not skip_missing:
                raise RuntimeError(f"Field {key} not found in {path}")
            continue
        arrays = data[key]
        if not isinstance(arrays, list):
            arrays = [arrays]
        domain.arrays_to_field(arrays, state.fields[key])


def extrap_quadh(u0, u1, u1p):
    '''
    Quadratic extrapolation from points 0, 1, 1.5 to point 2.
    Suffix `h` means half.
    '''
    u2 = (u0 - 6 * u1 + 8 * u1p) / 3
    return u2


def extrap_quad(u0, u1, u2):
    'Quadratic extrapolation from points 0, 1, 2 to point 3.'
    u3 = u0 - 3 * u1 + 3 * u2
    return u3


def extrap_linear(u0, u1):
    'Linear extrapolation from points 0, 1 to point 2.'
    u2 = 2 * u1 - u0
    return u2


def latin_hypercube(ndim, size, dtype):
    '''
    Returns `size` points from the unit cube.
    '''
    # XXX Copied from pyDOE.
    cut = np.linspace(0, 1, size + 1, dtype=dtype)
    u = np.random.rand(size, ndim).astype(dtype)
    a = cut[:size]
    b = cut[1:size + 1]
    rdpoints = np.zeros_like(u, dtype=dtype)
    for j in range(ndim):
        rdpoints[:, j] = u[:, j] * (b - a) + a
    H = np.zeros_like(rdpoints)
    for j in range(ndim):
        order = np.random.permutation(range(size))
        H[:, j] = rdpoints[order, j]
    return H


class Approx():

    def __init__(self, domain):
        self.domain = domain
        self.mod = domain.mod

    def stencil(self, q):
        'Returns: q, qxm, qxp, qym, qyp.'
        mod = self.mod
        st = [None] * 5
        st[0] = q
        st[1] = mod.roll(st[0], shift=1, axis=0)
        st[2] = mod.roll(st[0], shift=-1, axis=0)
        st[3] = mod.roll(st[0], shift=1, axis=1)
        st[4] = mod.roll(st[0], shift=-1, axis=1)
        return st

    def stencil5(self, st):
        'Returns: qxmm, qxpp, qymm, qypp.'
        mod = self.mod
        st5 = [None] * 4
        st5[0] = mod.roll(st[1], shift=1, axis=0)
        st5[1] = mod.roll(st[2], shift=-1, axis=0)
        st5[2] = mod.roll(st[3], shift=1, axis=1)
        st5[3] = mod.roll(st[4], shift=-1, axis=1)
        return st5

    def central(self, st):
        hx, hy = self.domain.step()
        q, qxm, qxp, qym, qyp = st
        q_x = (qxp - qxm) / (2 * hx)
        q_y = (qyp - qym) / (2 * hy)
        return q_x, q_y

    def apply_bc_extrap_linear(self, st):
        'Linear extrapolation from inner cells to halo cells.'
        domain = self.domain
        nx, ny = domain.size()
        ix, iy = domain.indices()
        mod = domain.mod
        extrap = extrap_quad
        st[1] = mod.where(ix == 0, extrap(st[2], st[0]), st[1])
        st[2] = mod.where(ix == nx - 1, extrap(st[1], st[0]), st[2])
        st[3] = mod.where(iy == 0, extrap(st[4], st[0]), st[3])
        st[4] = mod.where(iy == ny - 1, extrap(st[3], st[0]), st[4])
        return st

    def apply_bc_extrap_quad(self, st, st5):
        'Linear extrapolation from inner cells to halo cells.'
        domain = self.domain
        nx, ny = domain.size()
        ix, iy = domain.indices()
        mod = domain.mod
        extrap = extrap_quad
        st[1] = mod.where(ix == 0, extrap(st5[1], st[2], st[0]), st[1])
        st[2] = mod.where(ix == nx - 1, extrap(st5[0], st[1], st[0]), st[2])
        st[3] = mod.where(iy == 0, extrap(st5[3], st[4], st[0]), st[3])
        st[4] = mod.where(iy == ny - 1, extrap(st5[2], st[3], st[0]), st[4])
        return st

    def vorticity(self, u, v):
        u_st = self.stencil(u)
        v_st = self.stencil(v)
        self.apply_bc_extrap_quad(u_st, self.stencil5(u_st))
        self.apply_bc_extrap_quad(v_st, self.stencil5(v_st))
        _, u_y = self.central(u_st)
        v_x, _ = self.central(v_st)
        omega = v_x - u_y
        return omega


def struct_to_numpy(mod, d):
    if mod.is_tensor(d):
        return np.array(d)
    if isinstance(d, dict):
        for key in d:
            d[key] = struct_to_numpy(mod, d[key])
        return d
    if isinstance(d, list):
        return [struct_to_numpy(mod, a) for a in d]
    if isinstance(d, tuple):
        return tuple(struct_to_numpy(mod, a) for a in d)
    return d



#!/usr/bin/env python3

import argparse
import numpy as np
import scipy
import scipy.sparse as sp

import odil
from odil import printlog


def operator(ctx):
    mod = ctx.mod
    extra = ctx.extra
    args = extra.args

    res = []

    # Value in cell (i, j) taken from face (i-0.5, j).
    u_xm = ctx.field('ufx', 0, 0, loc='cc')
    # Value in cell (i, j) taken from face (i+0.5, j).
    u_xp = ctx.field('ufx', 1, 0, loc='cc')
    # Equation in cells, given derivative in x.
    hx = ctx.step('x')
    res += [(u_xp - u_xm) / hx - extra.ref['dudx']]

    # Boundary conditions at x=0.
    ufx = ctx.field('ufx')
    ixfx = ctx.indices('x', loc='nc')
    mask = mod.where(ixfx == 0, ctx.cast(1), ctx.cast(0))
    res += [(ufx - extra.ref['ufx']) * mask]

    # Average over two faces.
    uc = ctx.field('uc')
    res += [(u_xp + u_xm) * 0.5 - uc]

    # Non-grid array, will generate a full Jacobian.
    a = ctx.field('a')
    res += [a - extra.ref['a']]

    # Neural network.
    net_out = ctx.neural_net('net')(*extra.ref['net_in'])
    for i in range(args.Nnet):
        res += [(f'net{i}', net_out[i] - extra.ref['net_out'][i])]
    return res


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--Nx', type=int, default=3, help="Grid size in x")
    parser.add_argument('--Ny', type=int, default=2, help="Grid size in y")
    parser.add_argument('--Na', type=int, default=5, help="Size of array a")
    parser.add_argument('--Nnet',
                        type=int,
                        default=5,
                        help="Number of inputs and output in neural net")
    odil.util.add_arguments(parser)
    odil.linsolver.add_arguments(parser)
    parser.set_defaults(outdir='out_newton')
    parser.set_defaults(multigrid=0)
    parser.set_defaults(seed=1000)
    return parser.parse_args()


def make_problem(args):
    domain = odil.Domain(cshape=(args.Nx, args.Ny),
                         dimnames=['x', 'y'],
                         lower=(0, 0),
                         dtype=np.float64,
                         upper=(args.Nx, args.Ny),
                         multigrid=args.multigrid,
                         mg_interp=args.mg_interp,
                         mg_axes=[True, True],
                         mg_nlvl=args.nlvl)
    dtype = domain.dtype

    from odil import Field, Array
    state = odil.State(
        fields={
            'uc':
            Field(np.ones(domain.size(loc='cc')), loc='cc'),
            'ufx':
            Field(np.ones(domain.size(loc='nc')), loc='nc'),
            'a':
            Array(np.zeros(args.Na, dtype=dtype)),
            'net':
            domain.make_neural_net([args.Nnet, args.Nnet], activation='none'),
        })
    state = domain.init_state(state)

    def func(x, y):
        return 0.25 * x * y

    def func_x(x, y):
        return 0.25 * y

    extra = argparse.Namespace()
    xc, yc = domain.points(loc='cc')  # Cells.
    xn, yn = domain.points(loc='nn')  # Nodes.
    xfx, yfx = domain.points(loc='nc')  # Faces in x.
    xfy, yfy = domain.points(loc='cn')  # Faces in y.
    extra.ref = {
        'uc': func(xc, yc),
        'ufx': func(xfx, yfx),
        'dudx': func_x(xc, yc),
        'a': np.linspace(0, 1, state.fields['a'].array.shape[0], dtype=dtype),
    }
    # A linear neural net will transform one random matrix into another.
    # The number of degrees of freedom is `Nnet * Nnet + 1`
    # which includes the weights and biases.
    extra.ref['net_in'] = np.random.rand(args.Nnet, args.Nnet + 1)
    extra.ref['net_out'] = np.random.rand(args.Nnet, args.Nnet + 1)
    extra.args = args

    problem = odil.Problem(operator, domain, extra)
    return problem, state


def main():
    if odil.runtime.backend == 'jax':
        print('Skip test_newton.py with jax backend. Not implemented.')
        exit(0)
    args = parse_args()
    odil.setup_outdir(args)
    problem, state = make_problem(args)
    domain = problem.domain
    extra = problem.extra

    # Linearize the operator.
    vector, matrix = problem.linearize(state)
    vector = np.array(vector)

    # Solve the normal equations.
    delta = sp.linalg.spsolve(matrix.T @ matrix, -matrix.T @ vector)
    packed = domain.pack_state(state)
    domain.unpack_state(packed + delta, state)

    failed = 0
    for key in ['ufx', 'uc', 'a', 'net_out']:
        if key == 'net_out':
            value = domain.neural_net(state, 'net')(*extra.ref['net_in'])
        else:
            value = domain.field(state, key)
        error = value - extra.ref[key]
        error = np.sqrt(np.mean(np.square(error)))
        if error < 1e-6:
            status = 'PASS'
        else:
            status = 'FAIL'
            failed += 1
        print(f'{key:<4} {error:.6e} {status}')
    exit(failed)


if __name__ == "__main__":
    main()



import os, sys
import importlib.util


def lazy_import(name):
    spec = importlib.util.find_spec(name)
    loader = importlib.util.LazyLoader(spec.loader)
    spec.loader = loader
    module = importlib.util.module_from_spec(spec)
    sys.modules[name] = module
    loader.exec_module(module)
    return module


runtime = lazy_import("odil.runtime")
plot = lazy_import("odil.plot")

from .io import (
    parse_raw_xmf,
    read_raw,
    write_raw_xmf,
    write_raw_with_xmf,
    write_vtk_poly,
)
from .backend import ModBase, ModNumpy, ModTensorflow
from .history import History
from .util import (
    setup_outdir,
    optimize,
    make_callback,
    printlog,
    set_log_file,
)
from .core import (
    Domain,
    State,
    Problem,
    Field,
    MultigridField,
    NeuralNet,
    Array,
    restrict_to_coarser,
)
from . import linsolver
from . import core_min
from . import core
from .optimizer import EarlyStopError

del os, sys, importlib



#!/usr/bin/env python3

import argparse
import numpy as np
import pickle

import odil
from odil import plotutil
import matplotlib.pyplot as plt
from odil import printlog
from odil.core import Approx


def get_ref_u(name, args, domain):
    xw = domain.points()
    ndim = len(xw)
    if name == 'hat':
        p = 5
        u = np.prod([(1 - x) * x * 5 for x in xw], axis=0)
        u = (u**p / (1 + u**p))**(1 / p)
    elif name == 'osc':
        pi = np.pi
        k = args.osc_k
        x, y = xw
        u = np.sin(pi * (k * x)**2) * np.sin(pi * y)
    else:
        raise ValueError("Unknown name=" + name)
    return u


def get_ref_rhs(name, args, domain):
    xw = domain.points()
    ndim = len(xw)
    if name == 'osc':
        pi, cos, sin = np.pi, np.cos, np.sin
        k = args.osc_k
        x, y = xw
        fu = (((-4 * k**4 * pi**2 * x**2) - pi**2) * sin(k**2 * pi * x**2) +
              2 * k**2 * pi * cos(k**2 * pi * x**2)) * sin(pi * y)
    else:
        raise ValueError("Unknown name=" + name)
    return fu


def split_wm_wp(st, dirs):
    q = st[0]
    qwm = [st[2 * i + 1] for i in dirs]
    qwp = [st[2 * i + 2] for i in dirs]
    return q, qwm, qwp


def apply_bc_u_mod(st, iw, nw, dirs, mod):
    'Applies zero-Dirichlet boundary conditions.'
    q, qwm, qwp = split_wm_wp(st, dirs)
    zero = mod.cast(0, q.dtype)
    for i in dirs:
        extrap = odil.core.extrap_quadh
        qm = mod.where(iw[i] == 0, extrap(qwp[i], q, zero), qwm[i])
        qp = mod.where(iw[i] == nw[i] - 1, extrap(qwm[i], q, zero), qwp[i])
        qwm[i], qwp[i] = qm, qp
    for i in dirs:
        st[2 * i + 1] = qwm[i]
        st[2 * i + 2] = qwp[i]


def get_discrete_rhs(u, domain, mod):
    ndim = domain.ndim
    dirs = range(ndim)
    dw = domain.step()
    iw = domain.indices()
    nw = domain.size()
    u_st = [None] * (2 * ndim + 1)
    u_st[0] = u
    for i in dirs:
        u_st[2 * i + 1] = mod.roll(u, 1, i)
        u_st[2 * i + 2] = mod.roll(u, -1, i)
    apply_bc_u_mod(u_st, iw, nw, dirs, mod)
    u, uwm, uwp = split_wm_wp(u_st, dirs)
    u_ww = [(uwp[i] - 2 * u + uwm[i]) / dw[i]**2 for i in dirs]
    fu = sum(u_ww)
    return fu


def operator(ctx):
    domain = ctx.domain
    extra = ctx.extra
    args = extra.args
    mod = domain.mod
    ndim = domain.ndim
    dirs = range(ndim)
    dw = ctx.step()
    iw = ctx.indices()
    nw = ctx.size()

    def stencil_var(key):
        st = [ctx.field(key)]
        for i in dirs:
            w = [-1 if j == i else 0 for j in dirs]
            st.append(ctx.field(key, *w))
            w = [1 if j == i else 0 for j in dirs]
            st.append(ctx.field(key, *w))
        return st

    u_st = stencil_var('u')
    apply_bc_u_mod(u_st, iw, nw, dirs, mod=mod)
    u, uwm, uwp = split_wm_wp(u_st, dirs)
    u_ww = [(uwp[i] - 2 * u + uwm[i]) / dw[i]**2 for i in dirs]
    fu = sum(u_ww) - extra.rhs
    res = [fu]

    if args.mgloss:
        from functools import partial
        restrict = partial(odil.core.restrict_to_coarser,
                           loc='c' * ndim,
                           mod=mod)
        for _ in range(args.mgloss):
            fu = restrict(fu)
            res += [fu]
    return res


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--ndim', type=int, default=3, help="Space dimension")
    parser.add_argument('--N', type=int, default=32, help="Grid size")
    parser.add_argument('--cellbased',
                        type=int,
                        default=1,
                        help="Cell-based fields")
    parser.add_argument('--dump_xmf',
                        type=int,
                        default=0,
                        help="Dump XMF+RAW files")
    parser.add_argument('--plot', type=int, default=0, help="Enable plotting")
    parser.add_argument('--ref',
                        type=str,
                        default='hat',
                        choices=('hat', 'osc'),
                        help="Reference solution")
    parser.add_argument('--rhs',
                        type=str,
                        default='discrete',
                        choices=('discrete', 'exact'),
                        help="Reference right-hand side")
    parser.add_argument('--osc_k',
                        type=float,
                        default=2,
                        help="Parameter for ref='osc'")
    parser.add_argument('--mgloss',
                        type=int,
                        default=0,
                        help="Use multigrid norm in loss")
    odil.util.add_arguments(parser)
    odil.linsolver.add_arguments(parser)

    parser.set_defaults(frames=4,
                        report_every=100,
                        history_every=1,
                        plot_every=100)
    parser.set_defaults(history_full=50)

    parser.set_defaults(multigrid=1)
    parser.set_defaults(optimizer='lbfgsb')
    parser.set_defaults(every_factor=1)
    parser.set_defaults(lr=0.005)
    parser.set_defaults(double=1)

    parser.set_defaults(outdir='out_poisson')
    return parser.parse_args()


def write_field(u, name, path, domain, cellbased):
    ndim = domain.ndim
    dw = domain.step()
    axes = tuple(reversed(range(ndim)))
    u = np.transpose(u, axes)
    odil.write_raw_with_xmf(u, path, spacing=dw, name=name, cell=cellbased)


def plot_func(problem, state, epoch, frame, cbinfo):
    domain = problem.domain
    mod = domain.mod
    extra = problem.extra
    args = extra.args
    if args.frames == 0 and frame is not None:
        # Only plot the last frame.
        return
    ndim = domain.ndim
    key = 'u'
    paths = []
    suff = "" if frame is None else "_{:05d}".format(frame)
    if args.plot and ndim == 1:
        x = domain.points(0)
        fig, ax = plt.subplots()
        u = domain.field(state, key)
        ax.plot(x, u, label='epoch {:}'.format(epoch))
        ax.plot(x, extra.ref_u, label='reference')
        ax.set_ylim(0, 1)
        ax.set_xlabel('x')
        ax.set_ylabel('u')
        ax.legend(loc='upper left', bbox_to_anchor=(1., 1.))
        plotutil.savefig(fig, "u{}".format(suff))
        plt.close(fig)

        fig, ax = plt.subplots()
        ax.plot(x, extra.rhs)
        ax.set_xlabel('x')
        ax.set_ylabel('rhs')
        plotutil.savefig(fig, "rhs{}".format(suff))
        plt.close(fig)

    if args.dump_xmf and ndim in [2, 3]:
        u = domain.field(state, key)
        path = key + '{}.xmf'.format(suff)
        write_field(u, key, path, domain, args.cellbased)
        paths.append(path)

    if args.dump_data:
        x = domain.points()
        u = domain.field(state, key)
        path = "data{}.pickle".format(suff)
        d = dict()
        d['x'] = x
        d['u'] = u
        d['ref_u'] = extra.ref_u
        d['rhs'] = extra.rhs
        if 0 and args.multigrid:  # XXX
            d['u_cumsum'] = [q.numpy() for q in us]
        d = odil.core.struct_to_numpy(mod, d)
        with open(path, 'wb') as f:
            pickle.dump(d, f)
        paths.append(path)

    printlog(' '.join(paths))


def get_error(domain, extra, state, key):
    mod = domain.mod
    state_u = domain.field(state, key)
    du = state_u - extra.ref_u
    return np.sqrt(np.mean(du**2))


def history_func(problem, state, epoch, history, cbinfo):
    domain = problem.domain
    extra = problem.extra
    for key in state.fields:
        error = get_error(domain, extra, state, key)
        history.append('error_' + key, error)


def report_func(problem, state, epoch, cbinfo):
    res = dict()
    domain = problem.domain
    extra = problem.extra
    for key in state.fields:
        res[key] = get_error(domain, extra, state, key)
    printlog('error: ' + ', '.join('{}:{:.5g}'.format(*item)
                                   for item in res.items()))


def make_problem(args):
    dtype = np.float64 if args.double else np.float32
    ndim = args.ndim
    domain = odil.Domain(cshape=[args.N] * ndim,
                         dimnames=['x', 'y', 'z', 's'][:ndim],
                         multigrid=args.multigrid,
                         dtype=dtype)
    if domain.multigrid:
        printlog('multigrid levels:', domain.mg_cshapes)
    mod = domain.mod

    cellbased = args.cellbased
    if cellbased:
        xw = domain.points(loc='c' * ndim)
    else:
        xw = domain.points(loc='n' * ndim)

    # Reference solution.
    ref_u = get_ref_u(args.ref, args, domain)
    # Reference right-hand side.
    if args.rhs == 'discrete':
        rhs = get_discrete_rhs(ref_u, domain, mod)
    else:
        rhs = get_ref_rhs(args.ref, args, domain)

    # Initial state.
    state = odil.State()
    state.fields['u'] = None
    state = domain.init_state(state)

    extra = argparse.Namespace()
    extra.ref_u = ref_u
    extra.rhs = rhs
    extra.args = args
    if args.plot:
        write_field(extra.ref_u, 'u', 'ref_u.xmf', domain, cellbased)
        write_field(extra.rhs, 'rhs', 'rhs.xmf', domain, cellbased)

    problem = odil.Problem(operator, domain, extra)
    return problem, state


def main():
    args = parse_args()
    odil.setup_outdir(args)
    problem, state = make_problem(args)
    callback = odil.make_callback(problem,
                                  args,
                                  plot_func=plot_func,
                                  history_func=history_func,
                                  report_func=report_func)
    odil.util.optimize(args, args.optimizer, problem, state, callback)
    plot_func(problem, state, 0, None, None)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import numpy as np

import odil
from odil import printlog


def operator(ctx):
    res = []

    def func(x, y):
        return x * 0.25 + y * 0.5

    # Cells.
    xc, yc = ctx.points(loc='cc')
    uc = ctx.field('uc')
    res += [('uc', uc - func(xc, yc))]

    # Nodes.
    xn, yn = ctx.points(loc='nn')
    un = ctx.field('un')
    res += [('un', un - func(xn, yn))]

    # Faces in x.
    xfx, yfx = ctx.points(loc='nc')
    ufx = ctx.field('ufx')
    res += [('ufx', ufx - func(xfx, yfx))]

    # Faces in y.
    xfy, yfy = ctx.points(loc='cn')
    ufy = ctx.field('ufy')
    res += [('ufy', ufy - func(xfy, yfy))]
    return res


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--Nx', type=int, default=8, help="Grid size in x")
    parser.add_argument('--Ny', type=int, default=4, help="Grid size in y")
    parser.add_argument('--plot', type=int, default=1, help="Plot fields")
    odil.util.add_arguments(parser)
    odil.linsolver.add_arguments(parser)
    parser.set_defaults(outdir='out_fields')
    parser.set_defaults(multigrid=1)
    parser.set_defaults(optimizer='adamn', lr=1e-2)
    parser.set_defaults(frames=1,
                        plot_every=100,
                        report_every=10,
                        history_every=10)
    return parser.parse_args()


def plot(problem, state, epoch, frame, cbinfo=None):
    from odil import plotutil
    import matplotlib.pyplot as plt
    domain = problem.domain
    fig, ax = plt.subplots(figsize=(4, 2))
    kw = dict(vmin=0, vmax=1, cmap='Greys', clip_on=False, lw=0.5)

    # Cells.
    xc, yc = domain.points(loc='cc')
    uc = domain.field(state, 'uc')
    ax.scatter(xc, yc, s=10, c=uc, edgecolor='C0', label='uc', **kw)

    # Nodes.
    xn, yn = domain.points(loc='nn')
    un = domain.field(state, 'un')
    ax.scatter(xn, yn, s=10, c=un, edgecolor='C1', label='un', **kw)

    # Faces in x.
    xfx, yfx = domain.points(loc='nc')
    ufx = domain.field(state, 'ufx')
    ax.scatter(xfx, yfx, s=10, c=ufx, edgecolor='C2', label='ufx', **kw)

    # Faces in y.
    xfy, yfy = domain.points(loc='cn')
    ufy = domain.field(state, 'ufy')
    ax.scatter(xfy, yfy, s=10, c=ufy, edgecolor='C3', label='ufy', **kw)

    ax.legend(loc='lower left',
              bbox_to_anchor=(0.1, 1),
              ncol=4,
              handletextpad=0)

    ax.pcolormesh(xn,
                  yn,
                  uc,
                  edgecolor='k',
                  shading='flat',
                  zorder=0,
                  **dict(kw, lw=0.5))

    ax.set_aspect('equal')
    ax.set_axis_off()
    plotutil.savefig(fig, f'grid_{frame:05d}', printf=printlog)


def make_problem(args):
    dtype = np.float64 if args.double else np.float32
    domain = odil.Domain(cshape=(args.Nx, args.Ny),
                         dimnames=['x', 'y'],
                         lower=(0, 0),
                         upper=(2, 1),
                         dtype=dtype,
                         multigrid=args.multigrid,
                         mg_interp=args.mg_interp,
                         mg_axes=[True, True],
                         mg_nlvl=args.nlvl)
    xx, yy = domain.points('x', 'y', loc='cn')
    ixx, iyy = domain.indices('x', 'y', loc='cc')

    from odil import Field

    state = odil.State(
        fields={
            'uc': Field(np.zeros(domain.size(loc='cc')), loc='cc'),
            'un': Field(np.zeros(domain.size(loc='nn')), loc='nn'),
            'ufx': Field(np.zeros(domain.size(loc='nc')), loc='nc'),
            'ufy': Field(np.zeros(domain.size(loc='cn')), loc='cn'),
            'net': domain.make_neural_net([2, 4, 2]),
        })
    state = domain.init_state(state)
    problem = odil.Problem(operator, domain)
    return problem, state


def main():
    args = parse_args()
    odil.setup_outdir(args)
    problem, state = make_problem(args)
    callback = odil.make_callback(problem,
                                  args,
                                  plot_func=plot if args.plot else None)
    odil.util.optimize_grad(args, args.optimizer, problem, state, callback)


if __name__ == "__main__":
    main()



import os
import numpy as np

import matplotlib
if int(os.environ.get("ODIL_AGG", 1)):
    matplotlib.use('Agg')
import matplotlib.pyplot as plt
mplstyle = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                        'odil.mplstyle')
if int(os.environ.get("ODIL_STYLE", 1)):
    plt.style.use(mplstyle)

g_extlist = None


def set_extlist(extlist=None):
    global g_extlist
    if extlist is None:
        g_extlist = os.environ.get("ODIL_EXTLIST", 'png').split(',')
    else:
        g_extlist = extlist


set_extlist()


def apply_clip_box(ax, artists, lower=(0, 0), upper=(1, 1.02)):
    clipbox = matplotlib.transforms.TransformedBbox(
        matplotlib.transforms.Bbox([lower, upper]), ax.transAxes)
    for l in artists:
        l.set_clip_box(clipbox)


def savefig(fig,
            path,
            extlist=None,
            skip_existing=False,
            printf=print,
            **kwargs):
    if printf is None:
        printf = lambda _: None
    if extlist is None:
        extlist = g_extlist
    for ext in extlist:
        metadata = {
            'Date': None,
        } if ext == 'svg' else {
            'DateModified': None,
            'CreationDate': None,
        } if ext == 'pdf' else {}
        p = path + '.' + ext
        if skip_existing and os.path.isfile(p):
            printf("skip existing '{}'".format(p))
            continue
        printf(p)
        fig.savefig(p, metadata=metadata, **kwargs)


def savelegend(fig, ax, path, **kwargs):
    figleg, axleg = plt.subplots()
    handles, labels = ax.get_legend_handles_labels()
    legend = axleg.legend(handles, labels, loc='center', frameon=False)
    axleg.set_axis_off()
    figleg.canvas.draw()
    bbox = legend.get_window_extent().transformed(
        fig.dpi_scale_trans.inverted())
    savefig(figleg, path, bbox_inches=bbox, **kwargs)


def set_log_ticks(xaxis):
    locmin = matplotlib.ticker.LogLocator(base=10.,
                                          subs=np.arange(0.1, 0.99, 0.1),
                                          numticks=12)
    xaxis.set_minor_locator(locmin)
    xaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())



#!/usr/bin/env python3

import argparse
import numpy as np

import odil
from odil import printlog
from odil import plotutil
"""
Inference of unknown constant parameters in the advection-diffusion equation
from the solution at the initial and final time.
"""

C_DIFF = 0.01  # Diffusivity.
C_SRC = 0.1  # Uniform source.
C_VEL = 0.2  # Advection velocity.


def get_exact(t, x):
    '''
    Returns solution of the equation:
      u_t = nu * u_xx + src
    '''
    t = np.array(t)
    x = np.array(x)
    u = np.zeros_like(x)
    nu = C_DIFF
    xx = x - t * C_VEL
    ii = [1, 2, 3]
    for i in ii:
        k = 2 * i * np.pi
        u += np.cos(xx * k) * np.exp(-nu * k**2 * t)
    u /= 2 * len(ii)
    # Add uniform source.
    src = C_SRC
    u += src * t
    return u


def transform_u(u, extra, mod):
    u = mod.concatenate(
        [extra.u_init[None, :], u[1:-1], extra.u_final[None, :]], axis=0)
    return u


def operator_adv(ctx):
    mod = ctx.mod
    dt, dx = ctx.step('t', 'x')
    x = ctx.points('x')
    it, ix = ctx.indices('t', 'x')
    nt, nx = ctx.size('t', 'x')
    coeff = ctx.field('coeff')
    extra = ctx.extra

    def stencil_roll(q):
        return [
            mod.roll(q, shift=np.negative(s), axis=(0, 1))
            for s in [(0, 0), (0, -1), (0, 1), (-1, 0), (-1, -1), (-1, 1)]
        ]

    u = transform_u(ctx.field('u'), extra, mod)
    u_st = stencil_roll(u)
    u, uxm, uxp, um, umxm, umxp = u_st

    u_t = (u - um) / dt
    um_xx = (umxm - 2 * um + umxp) / (dx**2)
    u_xx = (uxm - 2 * u + uxp) / (dx**2)
    u_xx = 0.5 * (u_xx + um_xx)
    um_x = (um - umxm) / dx
    u_x = (u - uxm) / dx
    u_x = 0.5 * (u_x + um_x)

    # Discretization.
    fu = u_t - coeff[0] * u_xx - coeff[1] + coeff[2] * u_x
    res = [fu[1:]]
    return res


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--Nt', type=int, default=64, help="Grid size in t")
    parser.add_argument('--Nx', type=int, default=64, help="Grid size in x")
    odil.util.add_arguments(parser)
    odil.linsolver.add_arguments(parser)

    parser.set_defaults(plot_every=50, report_every=50, history_every=50)
    parser.set_defaults(frames=10)
    parser.set_defaults(optimizer='lbfgsb')
    parser.set_defaults(double=1)
    parser.set_defaults(multigrid=1)
    parser.set_defaults(outdir='out_infer_constant_adv')
    return parser.parse_args()


def plot(problem, state, epoch, frame, cbinfo=None):
    domain = problem.domain
    extra = problem.extra
    uu_exact = extra.uu_exact

    # Print current values of coefficients.
    printlog('diff={:.5g}, src={:.5g}, vel={:.5g}'.format(
        *np.array(state.fields['coeff'].array)))

    path0 = "u_{:05d}.png".format(frame)
    printlog(path0)

    uu = domain.field(state, 'u')
    uu = np.array(transform_u(uu, extra, domain.mod))

    umax = max(abs(np.max(uu_exact)), abs(np.min(uu_exact)))
    odil.plot.plot_1d(domain,
                      uu_exact,
                      uu,
                      path=path0,
                      cmap='RdBu_r',
                      nslices=5,
                      transpose=True,
                      transparent=False,
                      umin=-umax,
                      umax=umax)


def make_problem(args):
    dtype = np.float64 if args.double else np.float32
    domain = odil.Domain(cshape=(args.Nt, args.Nx),
                         dimnames=('t', 'x'),
                         lower=(0, -1),
                         upper=(1, 1),
                         dtype=dtype,
                         multigrid=args.multigrid,
                         mg_interp=args.mg_interp,
                         mg_nlvl=args.nlvl)

    # Evaluate exact solution, boundary and initial conditions.
    tt, xx = domain.points()
    # Node-based in time, cell-based in space.
    t1 = domain.points('t', loc='n.')
    x1 = domain.points('x', loc='.c')
    uu_exact = get_exact(tt, xx)
    u_init = get_exact(x1 * 0 + domain.lower[0], x1)
    u_final = get_exact(x1 * 0 + domain.upper[0], x1)

    # Initial state.
    state = odil.State(fields={
        'coeff': [0, 0, 0.001],
        'u': odil.Field(None, loc='nc'),
    })
    state = domain.init_state(state)

    extra = argparse.Namespace()
    extra.uu_exact = uu_exact
    extra.u_init = u_init
    extra.u_final = u_final
    extra.args = args

    problem = odil.Problem(operator_adv, domain, extra)
    return problem, state


def main():
    global problem, args

    args = parse_args()
    odil.setup_outdir(args)
    problem, state = make_problem(args)

    callback = odil.make_callback(problem, args, plot_func=plot)
    odil.optimize(args, args.optimizer, problem, state, callback)

    with open('done', 'w') as f:
        pass


if __name__ == "__main__":
    main()



import numpy as np
from argparse import Namespace


class Optimizer:

    def __init__(self, name=None, displayname=None, dtype=None):
        self.name = name
        self.displayname = displayname if displayname is not None else name
        self.dtype = dtype
        self.pinfo = None
        self.evals = 0

    def run(self,
            x0,
            loss_grad,
            epochs,
            callback=None,
            epoch_start=0,
            **kwargs):
        optinfo = Namespace()
        optinfo.evals = 0  # Number of `loss_grad()` evaluations.
        optinfo.epochs = 0  # Number of epochs actually done.
        return x0, optinfo


class EarlyStopError(Exception):

    def __init__(self, msg):
        super().__init__(msg)


class LbfgsbOptimizer(Optimizer):

    def __init__(self,
                 pgtol=1e-16,
                 m=50,
                 maxls=50,
                 factr=0,
                 dtype=None,
                 mod=None,
                 **kwargs):
        """
        pgtol: `float`
            Gradient convergence condition.
        m: `int`
            Maximum number of variable metric corrections used to define
            the limited memory matrix.
        maxls: `int`
            Maximum number of line search steps (per iteration).
        factr: `float`
            Convergence condition:
               1e12 (low accuracy),
               1e7 (moderate accuracy),
               10.0 (extremely high accuracy)
        """
        super().__init__(name="lbfgsb", displayname="L-BFGS-B", dtype=dtype)
        self.mod = mod
        self.pgtol = pgtol
        self.m = m
        self.maxls = maxls
        self.factr = factr
        self.evals = 0

    def run(self,
            x0,
            loss_grad,
            epochs=None,
            callback=None,
            epoch_start=0,
            **kwargs):
        self.epoch = epoch_start
        mod = self.mod
        dtype = self.dtype

        def split_by_sizes(array, sizes, axis=0):
            cumsum = np.cumsum(sizes)[:-1]
            return np.split(array, cumsum, axis=axis)

        def flat_to_arrays(x):
            x = np.array(x, dtype=dtype)
            sizes = [np.prod(a.shape) for a in x0]
            split = split_by_sizes(x[:sum(sizes)], sizes)
            arrays = [mod.reshape(s, a.shape) for s, a in zip(split, x0)]
            return arrays

        def arrays_to_flat(arrays):
            arrays = [np.array(a, dtype=np.float64) for a in arrays]
            x = np.concatenate([a.flatten() for a in arrays], axis=0)
            return x

        def callback_wrap(x):
            self.epoch += 1
            if callback:
                arrays = flat_to_arrays(x)
                callback(arrays, self.epoch, self.pinfo)

        def func_wrap(x):
            self.evals += 1
            arrays = flat_to_arrays(x)
            loss, grads, pinfo = loss_grad(arrays)
            loss = np.array(loss, dtype=np.float64)
            grad = arrays_to_flat(grads)
            self.pinfo = pinfo
            return loss, grad

        x0 = [np.array(a, dtype=np.float64) for a in x0]
        x0_flat = arrays_to_flat(x0)

        from scipy import optimize
        x, f, sinfo = optimize.fmin_l_bfgs_b(func=func_wrap,
                                             x0=x0_flat,
                                             maxiter=epochs,
                                             pgtol=self.pgtol,
                                             m=self.m,
                                             maxls=self.maxls,
                                             factr=self.factr,
                                             maxfun=np.inf,
                                             callback=callback_wrap)
        optinfo = Namespace()
        optinfo.warnflag = sinfo['warnflag']
        optinfo.task = sinfo['task']
        optinfo.evals = sinfo['funcalls']
        optinfo.epochs = sinfo['nit']
        if optinfo.warnflag not in [0, 1]:
            raise EarlyStopError(", ".join(
                "{:}={:}".format(key, sinfo.get(key, ''))
                for key in ['warnflag', 'task', 'funcalls', 'nit']))
        arrays = flat_to_arrays(x)
        return arrays, optinfo


class LbfgsOptimizer(Optimizer):

    def __init__(self,
                 pgtol=1e-16,
                 m=50,
                 maxls=50,
                 factr=0,
                 dtype=None,
                 mod=None,
                 **kwargs):
        """
        pgtol: `float`
            Gradient convergence condition.
        m: `int`
            Maximum number of variable metric corrections used to define
            the limited memory matrix.
        maxls: `int`
            Maximum number of line search steps (per iteration).
        factr: `float`
            Convergence condition:
               1e12 (low accuracy),
               1e7 (moderate accuracy),
               10.0 (extremely high accuracy)
        """
        super().__init__(name="lbfgs", displayname="L-BFGS_TF", dtype=dtype)
        self.mod = mod
        self.pgtol = pgtol
        self.m = m
        self.maxls = maxls
        self.factr = factr
        self.evals = 0
        self.last_x = None

    def run(self,
            x0,
            loss_grad,
            epochs=None,
            callback=None,
            epoch_start=0,
            **kwargs):
        self.epoch = epoch_start
        mod = self.mod
        tf = mod.tf
        import tensorflow_probability as tfp

        def flat_to_arrays(x):
            sizes = [np.prod(a.shape) for a in x0]
            split = mod.split_by_sizes(x[:sum(sizes)], sizes)
            arrays = [mod.reshape(s, a.shape) for s, a in zip(split, x0)]
            return arrays

        def arrays_to_flat(arrays):
            x = mod.concatenate([mod.flatten(a) for a in arrays], axis=0)
            return x

        def callback_wrap(x):
            self.epoch += 1
            if callback:
                sizes = [np.prod(a.shape) for a in x0]
                split = mod.split_by_sizes(x[:sum(sizes)], sizes)
                arrays = [mod.reshape(s, a.shape) for s, a in zip(split, x0)]
                callback(arrays, self.epoch, self.pinfo)

        def func_wrap(x):
            self.last_x = x
            self.evals += 1
            arrays = flat_to_arrays(x)
            loss, grads, pinfo = loss_grad(arrays)
            loss = mod.constant(loss)
            grad = arrays_to_flat(grads)
            self.pinfo = pinfo
            return loss, grad

        def stopping_condition(converged, failed):
            callback_wrap(self.last_x)
            if self.epoch > self.epochs:
                return tf.ones(converged.shape, dtype=bool)
            return tfp.optimizer.converged_all(converged, failed)

        x0_flat = arrays_to_flat(x0)
        self.last_x = x0_flat
        self.epochs = epochs
        res = tfp.optimizer.lbfgs_minimize(
            func_wrap,
            initial_position=x0_flat,
            max_iterations=epochs,
            num_correction_pairs=self.m,
            max_line_search_iterations=self.maxls,
            tolerance=-1,
            x_tolerance=-1,
            f_relative_tolerance=-1,
            stopping_condition=stopping_condition,
        )

        arrays = flat_to_arrays(res.position)
        optinfo = Namespace()
        optinfo.epochs = 0
        optinfo.evals = self.evals
        return arrays, optinfo


class AdamTfOptimizer(Optimizer):

    def __init__(self, dtype=None, mod=None, **kwargs):
        super().__init__(name="adam_tf", displayname="AdamTf", dtype=dtype)
        self.mod = mod

    def run(self,
            x0,
            loss_grad,
            epochs=None,
            callback=None,
            lr=1e-3,
            epoch_start=0,
            **kwargs):

        mod = self.mod
        tf = mod.tf

        class CustomModel(tf.keras.Model):

            def __init__(self, x):
                super().__init__()
                self.x = x
                self.evals = 0
                self.epoch = epoch_start

            def __call__(self):
                self.evals += 1
                return loss_grad(self.x)

            def train_step(self, _):
                loss, grads, pinfo = self()
                self.optimizer.apply_gradients(zip(grads, self.x))
                self.pinfo = pinfo
                return {'loss': loss}

        x = [mod.tf.Variable(e) for e in x0]
        model = CustomModel(x)

        class CustomCallback(tf.keras.callbacks.Callback):

            def on_epoch_end(self, epoch, logs=None):
                model.epoch = epoch_start + epoch
                if epoch > 0 and callback:
                    callback(model.x, model.epoch, model.pinfo)

        model.compile(optimizer=tf.keras.optimizers.Adam(learning_rate=lr),
                      run_eagerly=True)
        dummy = [1]  # Unused input.
        model.fit(dummy,
                  epochs=epochs + 1,
                  callbacks=[CustomCallback()],
                  verbose=0)
        optinfo = Namespace()
        optinfo.epochs = epochs
        optinfo.evals = model.evals
        return x, optinfo


class GdOptimizer(Optimizer):

    def __init__(self, dtype=None, mod=None, **kwargs):
        super().__init__(name="gd", displayname="GD", dtype=dtype)
        self.mod = mod

    def run(self,
            x0,
            loss_grad,
            epochs=None,
            callback=None,
            lr=1e-3,
            epoch_start=0,
            **kwargs):

        mod = self.mod
        x = [mod.copy(e) for e in x0]
        for epoch in range(epoch_start + 1, epoch_start + epochs + 1):
            self.evals += 1
            loss, grads, pinfo = loss_grad(x)
            for i in range(len(x)):
                x[i] -= grads[i] * lr
            if epoch > 0 and callback is not None:
                callback(x, epoch, pinfo)

        optinfo = Namespace()
        optinfo.epochs = epochs
        optinfo.evals = self.evals
        return x, optinfo


class AdamNativeOptimizer(Optimizer):

    def __init__(self, dtype=None, mod=None, **kwargs):
        super().__init__(name="adamn", displayname="AdamNative", dtype=dtype)
        self.mod = mod

    def run(self,
            x0,
            loss_grad,
            epochs=None,
            callback=None,
            lr=1e-3,
            epoch_start=0,
            beta_1=0.9,
            beta_2=0.999,
            epsilon=1e-7,
            jit=True,
            **kwargs):
        '''
        Based on
        https://github.com/keras-team/keras/blob/v2.12.0/keras/optimizers/adam.py
        '''

        dtype = self.dtype
        mod = self.mod
        lr = mod.cast(lr, dtype)
        beta_1 = mod.cast(beta_1, dtype)
        beta_2 = mod.cast(beta_2, dtype)

        def _step(x, m, v, grads, local_epoch):
            local_epoch = mod.cast(local_epoch, dtype)
            beta_1_power = beta_1**local_epoch
            beta_2_power = beta_2**local_epoch
            alpha = lr * mod.sqrt(1 - beta_2_power) / (1 - beta_1_power)
            m = [m + (g - m) * (1 - beta_1) for m, g in zip(m, grads)]
            v = [
                v + (mod.square(g) - v) * (1 - beta_2)
                for v, g in zip(v, grads)
            ]
            x = [
                x - (m * alpha) / (mod.sqrt(v) + epsilon)
                for x, m, v in zip(x, m, v)
            ]
            return x, m, v

        step = _step
        if jit:
            if mod.jax:
                step = mod.jax.jit(_step)
            elif mod.tf:
                step = mod.tf.function(_step, jit_compile=True)

        x = [mod.copy(e) for e in x0]
        m = [mod.zeros_like(e) for e in x0]
        v = [mod.zeros_like(e) for e in x0]
        for epoch in range(epoch_start + 1, epoch_start + epochs + 1):
            self.evals += 1
            loss, grads, pinfo = loss_grad(x)
            x, m, v = step(x, m, v, grads, mod.constant(epoch - epoch_start))
            if epoch > 0 and callback is not None:
                callback(x, epoch, pinfo)

        optinfo = Namespace()
        optinfo.epochs = epochs
        optinfo.evals = self.evals
        return x, optinfo


def make_optimizer(name, dtype=None, mod=None, **kwargs):
    if name == "lbfgsb":
        optimizer = LbfgsbOptimizer(dtype=dtype, mod=mod, **kwargs)
    elif name == "lbfgs":
        optimizer = LbfgsOptimizer(dtype=dtype, mod=mod, **kwargs)
    elif name == "adam_tf":
        optimizer = AdamTfOptimizer(dtype=dtype, mod=mod, **kwargs)
    elif name == "adam" or name == "adamn":
        optimizer = AdamNativeOptimizer(dtype=dtype, mod=mod, **kwargs)
    elif name == "gd":
        optimizer = GdOptimizer(dtype=dtype, mod=mod, **kwargs)
    else:
        raise ValueError("Unknown optimizer '{}'".format(name))
    return optimizer



'''
Wrappers for backend modules, e.g. TensorFlow, JAX, NumPy, CuPy.
The interface tends to follow NumPy's conventions.
'''

import numpy as np
from argparse import Namespace
from functools import partial


class ModBase:

    def __init__(self, mod=None):
        assert mod is not None
        self.mod = mod
        for name in [
                'int32',
                'float32',
                'float64',
                'linspace',
                'ones',
                'ones_like',
                'roll',
                'reshape',
                'stack',
                'abs',
                'cos',
                'sin',
                'exp',
                'zeros',
                'zeros_like',
                'square',
                'sqrt',
                'transpose',
                'minimum',
                'maximum',
                'meshgrid',
                'where',
        ]:
            setattr(self, name, getattr(mod, name))
        self.flatten = lambda x: mod.reshape(x, [-1])


class ModNumpy(ModBase):

    def __init__(self, mod=None, modsp=None, jax=None):
        mod = mod or np
        super().__init__(mod)
        self.cast = lambda x, dtype: mod.array(x, dtype=dtype)
        self.numpy = mod.array
        self.native = mod.array
        self.spnative = lambda x: x
        self.full = mod.full
        self.sum = mod.sum
        self.mean = mod.mean
        self.cumsum = mod.cumsum
        self.std = mod.std
        self.median = mod.median
        self.min = mod.min
        self.max = mod.max
        self.log = mod.log
        self.tanh = mod.tanh
        self.relu = lambda x: mod.maximum(x, 0)
        self.arctan2 = mod.arctan2
        self.sigmoid = lambda x: 1 / (1 + mod.exp(-x))
        self.arange = mod.arange
        self.norm = mod.linalg.norm
        self.solve = mod.linalg.solve
        self.mod = mod
        self.modsp = modsp
        self.moveaxis = mod.moveaxis
        self.hstack = mod.hstack
        self.concatenate = mod.concatenate
        self.ndarray = mod.ndarray
        self.constant = mod.array
        self.is_tensor = lambda x: isinstance(x, mod.ndarray)
        self.clip = mod.clip
        self.floor = mod.floor
        self.copy = mod.copy
        self.gather_nd = lambda u, idx: u[tuple(mod.moveaxis(idx, -1, 0))]
        self.pad = mod.pad
        self.einsum = mod.einsum
        self.batch_to_space = None
        self.jax = jax
        self.tf = None
        self.array = mod.array
        self.matmul = mod.matmul
        self.variable = mod.array
        if jax:

            def split_by_sizes(array, sizes, axis=0):
                cumsum = np.cumsum(sizes)[:-1]
                return jax.numpy.split(array, cumsum, axis=axis)
        else:

            def split_by_sizes(array, sizes, axis=0):
                cumsum = np.cumsum(sizes)[:-1]
                return mod.split(array, cumsum, axis=axis)

        self.split_by_sizes = split_by_sizes

        if jax:
            self.stop_gradient = jax.lax.stop_gradient
            self.broadcast_to = jax.numpy.broadcast_to
            self.jit_wrap = lambda **kwargs: partial(jax.jit, **kwargs)
        else:
            self.broadcast_to = mod.broadcast_to

        def convolution(input, filters, strides, padding):
            '''
            input: n-dimensional input array
            filters: n-dimiensional kernel
            strides: `int` or sequence of n integers
            '''
            if isinstance(strides, int):
                strides = (strides, ) * len(input.shape)
            input = mod.reshape(input, (1, 1) + input.shape)
            filters = mod.reshape(filters, (1, 1) + filters.shape)
            res = jax.lax.conv(lhs=input,
                               rhs=filters,
                               window_strides=strides,
                               padding=padding)
            res = res[0, 0, ...]
            return res

        self.convolution = convolution

        # Random.
        self.random = Namespace()
        if jax:
            self.random.key = None

            def set_seed(seed):
                self.random.key = jax.random.PRNGKey(seed)

            def uniform(shape, minval, maxval, dtype):
                if self.random.key is None:
                    set_seed(np.random.default_rng().integers(1 << 16))
                self.random.key, subkey = jax.random.split(self.random.key)
                return jax.random.uniform(subkey,
                                          shape=shape,
                                          minval=minval,
                                          maxval=maxval,
                                          dtype=dtype)

            def normal(shape, mean=0, stddev=1, dtype=None):
                if self.random.key is None:
                    set_seed(np.random.default_rng().integers(1 << 16))
                self.random.key, subkey = jax.random.split(self.random.key)
                mean = self.cast(mean, dtype)
                stddev = self.cast(stddev, dtype)
                return mean + stddev * jax.random.normal(
                    subkey, shape=shape, dtype=dtype)
        else:

            def set_seed(seed):
                self.random.set_seed(seed)

            def uniform(shape, minval, maxval, dtype):
                return mod.random.uniform(low=minval, high=maxval,
                                          size=shape).astype(dtype)

            def normal(shape, mean, stddev, dtype):
                return mod.random.normal(loc=mean, scale=stddev,
                                         size=shape).astype(dtype)

        self.random.set_seed = set_seed
        self.random.uniform = uniform
        self.random.normal = normal

        def conv_transpose(input,
                           filters,
                           output_shape=None,
                           strides=None,
                           padding=None):
            dim = len(input.shape)
            if isinstance(strides, int):
                strides = (strides, ) * dim
            res = jax.lax.conv_transpose(lhs=input,
                                         rhs=filters,
                                         strides=strides,
                                         padding=padding)
            return res

        self.conv_transpose = conv_transpose
        if modsp is not None:
            self.csr_matrix = modsp.csr_matrix
            self.diags = modsp.diags
            self.bmat = modsp.bmat
            self.block_diag = modsp.block_diag
            self.tril = modsp.tril
            tmp = self.numpy
            self.numpy = (lambda x: x
                          if isinstance(x, modsp.csr_matrix) else tmp(x))
            self.spnorm = modsp.linalg.norm
            self.spsolve = modsp.linalg.spsolve


class ModTensorflow(ModBase):

    def __init__(self, mod=None, modsp=None):
        super().__init__(mod)
        self.batch_to_space = mod.batch_to_space
        self.cast = mod.cast
        self.numpy = lambda x: x.numpy()
        self.native = lambda x: x
        self.spnative = lambda x: x.numpy() if hasattr(x, 'numpy') else x
        self.sum = mod.reduce_sum
        self.mean = mod.reduce_mean
        self.max = mod.reduce_max
        self.min = mod.reduce_min
        self.std = mod.math.reduce_std
        self.log = mod.math.log
        self.cumsum = mod.math.cumsum
        self.tanh = mod.math.tanh
        self.relu = mod.nn.relu
        self.sigmoid = mod.math.sigmoid
        self.full = mod.fill
        self.arange = mod.range
        self.norm = lambda x: mod.reduce_sum(x**2)**0.5
        self.mod = mod
        self.modsp = modsp
        self.floor = mod.math.floor
        self.arctan2 = mod.math.atan2
        self.concatenate = mod.concat
        self.constant = mod.constant
        self.is_tensor = mod.is_tensor
        self.clip = mod.clip_by_value
        self.copy = mod.identity
        self.gather_nd = mod.gather_nd
        self.batch_to_space = mod.batch_to_space
        self.conv_transpose = mod.nn.conv_transpose
        self.random = mod.random
        self.split_by_sizes = mod.split
        self.einsum = None
        self.array = mod.constant
        self.matmul = mod.linalg.matmul
        self.variable = mod.Variable
        self.tf = mod
        self.stop_gradient = mod.stop_gradient
        self.broadcast_to = mod.broadcast_to

        def jit_wrap(static_argnames=None, **kwargs):
            return partial(mod.function, **kwargs)

        self.jit_wrap = jit_wrap
        self.jax = None

        def pad(array, pad_width, mode):
            return mod.pad(array, paddings=pad_width, mode=mode)

        self.pad = pad

        def convolution(input, filters, strides, padding):
            '''
            input: n-dimensional input array
            filters: n-dimiensional kernel
            strides: `int` or sequence of n integers
            '''
            input = mod.reshape(input, (1, ) + input.shape + (1, ))
            filters = mod.reshape(filters, filters.shape + (1, 1))
            res = mod.nn.convolution(input=input,
                                     filters=filters,
                                     strides=strides,
                                     padding=padding)
            res = res[0, ..., 0]
            return res

        self.convolution = convolution

        # Random.
        self.random = Namespace()

        def set_seed(seed):
            mod.random.set_seed(seed)

        self.random.set_seed = set_seed
        self.random.uniform = mod.random.uniform
        self.random.normal = mod.random.normal

        if modsp is not None:

            def csr_matrix(*args, **kwargs):
                if 'dtype' in kwargs:
                    kwargs['dtype'] = {
                        mod.float32: np.float32,
                        mod.float64: np.float64
                    }[kwargs['dtype']]
                return modsp.csr_matrix(*args, **kwargs)

            self.csr_matrix = csr_matrix
            self.diags = modsp.diags
            self.bmat = modsp.bmat
            self.block_diag = modsp.block_diag
            self.tril = modsp.tril
            tmp = self.numpy
            self.numpy = (lambda x: x
                          if isinstance(x, modsp.csr_matrix) else tmp(x))
            self.spnorm = modsp.linalg.norm
            self.spsolve = modsp.linalg.spsolve


class ModCupy(ModBase):

    def __init__(self, mod=None, modsp=None):
        super().__init__(mod)
        self.cast = lambda x, dtype: mod.array(x, dtype=dtype)
        self.numpy = lambda x: x.get() if isinstance(x, mod.ndarray) else x
        self.full = mod.full
        self.sum = mod.sum
        self.native = mod.array
        self.spnative = lambda x: x
        self.arange = mod.arange
        self.norm = mod.linalg.norm
        self.solve = mod.linalg.solve
        self.mod = mod
        self.modsp = modsp
        self.moveaxis = mod.moveaxis
        self.hstack = mod.hstack
        if modsp is not None:
            self.csr_matrix = modsp.csr_matrix
            self.diags = modsp.diags
            self.bmat = modsp.bmat

            def block_diag(bb):
                n = len(bb)
                res = [[None for _ in range(n)] for _ in range(n)]
                for i in range(n):
                    for j in range(n):
                        res[i][i] = bb[i]
                return modsp.bmat(res, format='csr')

            self.block_diag = block_diag
            self.tril = modsp.tril
            tmp = self.numpy
            self.numpy = (lambda x: x.get()
                          if isinstance(x, modsp.csr_matrix) else tmp(x))
            self.spnorm = modsp.linalg.norm
            self.spsolve = modsp.linalg.spsolve



import os
import sys
import numpy as np

from .backend import ModNumpy, ModTensorflow

if not int(os.environ.get("ODIL_MT", 0)):
    os.environ["OMP_NUM_THREADS"] = "1"
    os.environ["XLA_FLAGS"] = ("--xla_cpu_multi_thread_eigen=false "
                               "intra_op_parallelism_threads=1")
    os.environ["TENSORFLOW_INTER_OP_PARALLELISM"] = "1"
    os.environ["TENSORFLOW_INTRA_OP_PARALLELISM"] = "1"

usegpu = os.environ.get("CUDA_VISIBLE_DEVICES", "") not in ["", "-1"]
if not usegpu:
    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

if not int(os.environ.get("ODIL_WARN", 0)):
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
    import warnings
    warnings.simplefilter(action='ignore', category=FutureWarning)
    warnings.simplefilter(action='ignore', category=Warning)

backend = os.environ.get("ODIL_BACKEND", 'tf')
jit = bool(int(os.environ.get("ODIL_JIT", 0)))
if backend == 'tf':
    import tensorflow as tf
    tf.config.threading.set_inter_op_parallelism_threads(1)
    tf.config.threading.set_intra_op_parallelism_threads(1)
    gpus = tf.config.list_physical_devices('GPU')
    if gpus:
        try:
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.list_logical_devices('GPU')
        except RuntimeError as e:
            print(e)

    jax = None
    mod = ModTensorflow(tf)
elif backend == 'jax':
    tf = None
    import jax
    if not usegpu:
        jax.config.update('jax_platform_name', 'cpu')
    # Enable support for float64 operations.
    jax.config.update("jax_enable_x64", True)

    mod = ModNumpy(jax.numpy, jax=jax)
else:
    sys.stderr.write(
        f"Unknown ODIL_BACKEND='{backend}', options are: tf, jax\n")
    exit(1)

# Default data type.
dtype_name = os.environ.get("ODIL_DTYPE", 'float32')
if dtype_name in ['float32', 'float64']:
    dtype = np.dtype(dtype_name)
else:
    sys.stderr.write(
        f"Expected ODIL_DTYPE=float32 or float64, got '{dtype}' \n")
    exit(1)



#!/usr/bin/env python3

import numpy as np
from copy import deepcopy
import odil
from odil.runtime import mod

cases = ['pack', 'arrays']


def test(case, dim):
    cshape = (1 + np.array(range(dim))) * 2
    dimnames = ['x', 'y', 'z', 'w'][:dim]
    domain = odil.Domain(cshape=cshape,
                         dimnames=dimnames,
                         multigrid=1,
                         mg_convert_all=False)

    state = odil.State(
        fields={
            'field': np.random.rand(*cshape),
            'mgfield': domain.regular_to_multigrid(  #
                np.random.rand(*cshape)),
            'net': domain.make_neural_net([3, 3]),
            'array': [1, 2, 3],
        })
    state = domain.init_state(state)
    state2 = deepcopy(state)

    def upd(u):
        return u + 1

    if case == cases[0]:
        # Modify state through packed array.
        packed = domain.pack_state(state)
        packed = upd(packed)
        domain.unpack_state(packed, state)
    elif case == cases[1]:
        # Modify state through raw arrays.
        arrays = domain.arrays_from_state(state)
        for i in range(len(arrays)):
            arrays[i] = upd(arrays[i])
        domain.arrays_to_state(arrays, state)

    # Modify state directly.
    for f in state2.fields.values():
        if isinstance(f, odil.core.Field):
            f.array = upd(f.array)
        elif isinstance(f, odil.core.MultigridField):
            for t in f.terms:
                t.array = upd(t.array)
        elif isinstance(f, odil.core.NeuralNet):
            for i in range(len(f.weights)):
                f.weights[i] = upd(f.weights[i])
                f.biases[i] = upd(f.biases[i])
        elif isinstance(f, odil.core.Array):
            f.array = upd(f.array)

    # Check that both coincide.
    error = max(domain.pack_state(state) - domain.pack_state(state2))
    return error


def main():
    failed = 0
    for case in cases:
        for dim in [1, 2]:
            try:
                error = test(case, dim)
            except Exception as e:
                error = e
                raise e
            failed += bool(error)
            print(f'case={case:>6} dim={dim}:',
                  f"FAIL {error}" if error else "PASS")
    exit(failed)


if __name__ == "__main__":
    main()



import os
import re
import numpy as np


def parse_raw_xmf(xmfpath):
    '''
    Returns shape and path to `.raw` file
    xmfpath: path to `.xmf` metadata file
    '''
    with open(xmfpath) as fin:
        text = ''.join(fin.read().split('\n'))
    m = re.findall(
        '<Xdmf.*<Attribute.*'
        '<DataItem.*<DataItem.*'
        '<DataItem.*Dimensions="(\d*) (\d*) (\d*)".*?> *([a-z0-9_.]*)',
        text)[0]
    shape = tuple(map(int, m[:3]))
    rawpath = m[3]
    rawpath = os.path.join(os.path.dirname(xmfpath), rawpath)
    return shape, rawpath


def read_raw(xmfpath):
    '''
    Returns array from scalar field in raw format.
    xmfpath: path to xmf metadata file
    '''
    shape, rawpath = parse_raw_xmf(xmfpath)
    u = np.fromfile(rawpath).reshape(shape)
    return u


def write_raw_xmf(xmfpath,
                  rawpath,
                  count,
                  spacing=(1, 1, 1),
                  name='data',
                  precision=8,
                  cell=True):
    '''
    Writes XMF metadata for a `.raw` datafile.
    xmfpath: path to output `.xmf` file
    rawpath: path to binary `.raw` file to be linked
    count: array size as (Nz, Ny, Nx)
    name: name of field
    cell: cell-centered field, else node-centered
    '''

    txt = '''\
<?xml version="1.0" ?>
<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>
<Xdmf Version="2.0">
 <Domain>
   <Grid Name="mesh" GridType="Uniform">
     <Topology TopologyType="{dim}DCORECTMesh" Dimensions="{nodes*}"/>
     <Geometry GeometryType="{geomtype}">
       <DataItem Name="Origin" Dimensions="{dim}" NumberType="Float" Precision="8" Format="XML">
         {origin*}
       </DataItem>
       <DataItem Name="Spacing" Dimensions="{dim}" NumberType="Float" Precision="8" Format="XML">
         {spacing*}
       </DataItem>
     </Geometry>
     <Attribute Name="{name}" AttributeType="Scalar" Center="{center}">
       <DataItem ItemType="HyperSlab" Dimensions="{countd*}" Type="HyperSlab">
           <DataItem Dimensions="3 {dim}" Format="XML">
             {start*}
             {stride*}
             {count*}
           </DataItem>
           <DataItem Dimensions="{bindim*}" Seek="{seek}" Precision="{precision}" NumberType="{type}" Format="Binary">
             {binpath}
           </DataItem>
       </DataItem>
     </Attribute>
   </Grid>
 </Domain>
</Xdmf>
'''

    def tostrrev(v):
        return ' '.join(map(str, reversed(v)))

    def tostr(v):
        return ' '.join(map(str, v))

    info = dict()
    dim = 3
    info['name'] = name
    info['dim'] = dim
    info['origin'] = tostrrev([0] * dim)
    info['spacing'] = tostrrev(spacing)
    info['start'] = tostrrev([0] * dim)
    info['stride'] = tostrrev([1] * dim)
    info['count'] = tostr(count)
    info['bindim'] = tostr(count)
    info['countd'] = tostr(count)
    if cell:
        info['nodes'] = tostr([a + 1 for a in count])
        info['center'] = 'Cell'
    else:
        info['nodes'] = tostr([a for a in count])
        info['center'] = 'Node'
    info['precision'] = precision
    if precision == 8:
        info['type'] = 'Double'
    else:
        info['type'] = 'Float'
    info['binpath'] = rawpath
    info['seek'] = '0'
    info['geomtype'] = 'ORIGIN_DXDYDZ'
    # Remove '*' which are only used in `aphros/src/dump/xmf.ipp`.
    txt = txt.replace('*}', '}')
    txt = txt.format(**info)

    with open(xmfpath, 'w') as fout:
        fout.write(txt)


def write_raw_with_xmf(u,
                       xmfpath,
                       rawpath=None,
                       spacing=(1, 1, 1),
                       cell=True,
                       name='data'):
    '''
    Writes binary data in raw format with XMF metadata.
    u: np.ndarray to write, shape (Nz, Ny, Nx)
    spacing: cell size
    name: name of field
    '''
    if len(u.shape) != 3:
        u = u.reshape((1, ) + u.shape)
    if len(spacing) != 3:
        spacing = list(spacing) + [min(spacing)]
    precision = 4 if u.dtype == np.float32 else 8
    if rawpath is None:
        rawpath = os.path.splitext(xmfpath)[0] + ".raw"
    rawrelpath = os.path.relpath(rawpath, start=os.path.dirname(xmfpath))
    write_raw_xmf(xmfpath, rawrelpath, u.shape, spacing, name, precision, cell)
    u.tofile(rawpath)
    return xmfpath


def write_vtk_poly(fout,
                   points,
                   polygons,
                   point_fields=None,
                   cell_fields=None,
                   comment="",
                   fmt='%.16g'):
    """
    Writes polygons to ASCII legacy VTK file.
    fout: `str` or file-like
        Path to output legacy VTK file or file-like object.
    points: `numpy.ndarray`, shape (npoints, 3)
        3D points.
    polygons: `list` [`list` [ `int` ]], shape (ncells, ...)
        Polygons as lists of indices in `points`.
    point_fields: `dict` from `str` to array of shape (npoints,)
        Mapping from names to arrays storing scalar fields on points.
    cell_fields: `dict` from `str` to array of shape (ncells,)
        Mapping from names to arrays storing scalar fields on cells.
    """
    path = None
    if type(fout) is str:
        path = fout
        fout = open(path, 'wb')

    def write(data):
        if type(data) is str:
            data = data.encode()
        fout.write(data)

    write("# vtk DataFile Version 2.0\n")

    write(comment + '\n')

    write("ASCII\n")

    write("DATASET POLYDATA\n")

    # Write points.
    npoints = len(points)
    write("POINTS {:} float\n".format(npoints))
    np.savetxt(fout, points, fmt=fmt)

    # Write cells.
    ncells = len(polygons)
    num_poly_data = len(polygons) + sum([len(p) for p in polygons])
    write("POLYGONS {:} {:}\n".format(ncells, num_poly_data))
    for p in polygons:
        write(' '.join(map(str, [len(p)] + p)) + '\n')

    # Write point fields.
    if point_fields is None:
        point_fields = dict()
    if point_fields:
        write("\nPOINT_DATA {:}\n".format(npoints))
    for name, array in point_fields.items():
        array = np.reshape(array, -1)
        if array.size != npoints:
            raise RuntimeError(
                f"Expected equal array.size={array.size} and npoints={npoints}"
            )
        write("SCALARS {:} float\n".format(name))
        write("LOOKUP_TABLE default\n")
        np.savetxt(fout, array, fmt=fmt)

    # Write cell fields.
    if cell_fields is None:
        cell_fields = dict()
    if cell_fields:
        write("\nCELL_DATA {:}\n".format(ncells))
    for name, array in cell_fields.items():
        array = np.reshape(array, -1)
        if array.size != ncells:
            raise RuntimeError(
                f"Expected equal array.size={array.size} and ncells={ncells}")
        write("SCALARS {:} float\n".format(name))
        write("LOOKUP_TABLE default\n")
        np.savetxt(fout, array, fmt=fmt)

    if path:
        fout.close()



import os
import functools


def cache_to_file(targetbase, update=False, name_with_arg0=False):
    """
    Factory for a decorator that caches the result of
    function and stores it to a target file.

    targetbase: base path to cache file
    update: force cache update
    name_with_arg0: append cache name by first argument converted to string

    Example: Creates file "_cache_7.pickle" with `int(7)`.

    @cache_to_file("_cache.pickle", name_with_arg0=False)
    def F(a):
        return a
    F(7)
    """

    ext = os.path.splitext(targetbase)[1]
    if ext == '.pickle':
        import pickle

        def load(path):
            with open(path, 'rb') as f:
                print("Loading cache '{}'".format(path))
                return pickle.load(f)

        def save(content, path):
            with open(path, 'wb') as f:
                print("Saving cache '{}'".format(path))
                pickle.dump(content, f, protocol=4)
    elif ext == '.json':
        import json

        def load(path):
            with open(path, 'r') as f:
                print("Loading cache '{}'".format(path))
                return json.load(f)

        def save(content, path):
            with open(path, 'w') as f:
                print("Saving cache '{}'".format(path))
                json.dump(content, f)

    elif ext == '.npy':
        import numpy

        def load(path):
            print("Loading cache '{}'".format(path))
            return numpy.load(path, allow_pickle=True).item()

        def save(content, path):
            print("Saving cache '{}'".format(path))
            numpy.save(path, content, allow_pickle=True)
    else:
        raise ValueError(
            "Unrecognized extension '{}', expected .pickle.".format(ext))

    def decorator(func):

        def inner(*args, **kwargs):
            name, ext = os.path.splitext(targetbase)
            if len(args) and name_with_arg0:
                name += '_{:}'.format(args[0])
            target = name + ext
            if not update and os.path.isfile(target):
                return load(target)
            result = func(*args, **kwargs)
            d = os.path.dirname(target)
            if d: os.makedirs(d, exist_ok=True)
            save(result, target)
            return result

        return functools.wraps(func)(inner)

    return decorator



#!/usr/bin/env python3

import argparse
import numpy as np

import odil
from odil import printlog


def operator(ctx):
    extra = ctx.extra

    res = []

    for key in ['uc', 'un', 'ufx', 'ufy']:
        res += [(key, ctx.field(key) - extra.ref[key])]
    res += [('a', ctx.field('a') - extra.ref['a'])]
    net_a = ctx.neural_net('net')(ctx.field('a'))[0]
    res += [('net_a', net_a - extra.ref['net_a'])]
    return res


def parse_args():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--Nx', type=int, default=8, help="Grid size in x")
    parser.add_argument('--Ny', type=int, default=4, help="Grid size in y")
    parser.add_argument('--optimizers',
                        type=str,
                        default='lbfgsb,adamn',
                        help="Optimizers to test, comma-separated list")
    odil.util.add_arguments(parser)
    odil.linsolver.add_arguments(parser)
    parser.set_defaults(outdir='out_optimize')
    parser.set_defaults(multigrid=1)
    parser.set_defaults(lr=0.1)
    parser.set_defaults(seed=1)
    parser.set_defaults(epochs=1000, report_every=100, history_every=100)
    return parser.parse_args()


def make_problem(args):
    domain = odil.Domain(cshape=(args.Nx, args.Ny),
                         dimnames=['x', 'y'],
                         lower=(0, 0),
                         upper=(2, 1),
                         multigrid=args.multigrid,
                         mg_interp=args.mg_interp,
                         mg_axes=[True, True],
                         mg_nlvl=args.nlvl)
    mod = domain.mod
    dtype = domain.dtype

    xx, yy = domain.points('x', 'y', loc='cn')
    ixx, iyy = domain.indices('x', 'y', loc='cc')

    from odil import Field, Array

    state = odil.State(
        fields={
            'uc': Field(np.zeros(domain.size(loc='cc')), loc='cc'),
            'un': Field(np.zeros(domain.size(loc='nn')), loc='nn'),
            'ufx': Field(np.zeros(domain.size(loc='nc')), loc='nc'),
            'ufy': Field(np.zeros(domain.size(loc='cn')), loc='cn'),
            'a': Array(np.zeros(5)),
            'net': domain.make_neural_net([1, 7, 1]),
        })
    state = domain.init_state(state)

    def func(x, y):
        return x * 0.25 + y * 0.5

    extra = argparse.Namespace()
    xc, yc = domain.points(loc='cc')  # Cells.
    xn, yn = domain.points(loc='nn')  # Nodes.
    xfx, yfx = domain.points(loc='nc')  # Faces in x.
    xfy, yfy = domain.points(loc='cn')  # Faces in y.
    extra.ref = {
        'uc': func(xc, yc),
        'un': func(xn, yn),
        'ufx': func(xfx, yfx),
        'ufy': func(xfy, yfy),
        'a': np.arange(state.fields['a'].array.shape[0], dtype=dtype),
    }
    extra.ref['net_a'] = extra.ref['a'] * 0.5

    problem = odil.Problem(operator, domain, extra)
    return problem, state


def main():
    args = parse_args()
    odil.setup_outdir(args)
    failed = 0
    for opt in args.optimizers.split(','):
        problem, state = make_problem(args)
        domain = problem.domain
        extra = problem.extra
        callback = odil.make_callback(problem, args)
        odil.util.optimize_grad(args, opt, problem, state, callback)
        error = [
            domain.field(state, key) - extra.ref[key]
            for key in ['uc', 'un', 'ufx', 'ufy', 'a']
        ]
        error.append(
            domain.neural_net(state, 'net')(domain.field(state, 'a')) -
            extra.ref['net_a'])
        error = np.sqrt(sum(np.mean(np.square(e)) for e in error))
        if error < 1e-2:
            status = 'PASS'
        else:
            status = 'FAIL'
            failed += 1
        print(f'opt={opt:>6} {status} {error:.8g}')
    exit(failed)


if __name__ == "__main__":
    main()
